from __future__ import absolute_import

import select

from .. import io, scheduler, util


def green_select(rlist, wlist, xlist, timeout=None):
    robjs = {}
    wobjs = {}
    fds = {}

    for fd in rlist:
        fdnum = fd if isinstance(fd, int) else fd.fileno()
        robjs[fdnum] = fd
        fds[fdnum] = 1

    for fd in wlist:
        fdnum = fd if isinstance(fd, int) else fd.fileno()
        wobjs[fdnum] = fd
        if fdnum in fds:
            fds[fdnum] |= 2
        else:
            fds[fdnum] = 2

    events = io.wait_fds(fds.items(), timeout=timeout, inmask=1, outmask=2)

    rlist_out, wlist_out = [], []
    for fd, event in events:
        if event & 1:
            rlist_out.append(robjs[fd])
        if event & 2:
            wlist_out.append(wobjs[fd])

    return rlist_out, wlist_out, []


if hasattr(select, "poll"):
    all_poll_evs = select.POLLIN | select.POLLOUT | select.POLLPRI
else:
    all_poll_evs = 0


class green_poll(object):
    def __init__(self):
        self._registry = {}

    def modify(self, fd, eventmask):
        fd = fd if isinstance(fd, int) else fd.fileno()
        if fd not in self._registry:
            raise IOError(2, "No such file or directory")
        self._registry[fd] = eventmask

    def poll(self, timeout=None):
        if timeout is not None:
            timeout = float(timeout) / 1000
        return io.wait_fds(self._registry.items(), timeout=timeout,
                           inmask=select.POLLIN, outmask=select.POLLOUT)

    def register(self, fd, eventmask=all_poll_evs):
        fd = fd if isinstance(fd, int) else fd.fileno()
        self._registry[fd] = eventmask

    def unregister(self, fd):
        fd = fd if isinstance(fd, int) else fd.fileno()
        del self._registry[fd]


if hasattr(select, 'epoll'):
    all_epoll_evs = select.EPOLLIN | select.EPOLLOUT | select.EPOLLPRI
else:
    all_epoll_evs = 0


class green_epoll(object):
    def __init__(self, sizehint=-1, from_ep=None):
        self._readable = util.Event()
        if from_ep:
            self._epoll = from_ep
        else:
            self._epoll = original_epoll(sizehint)

    def _on_readable(self):
        self._readable.set()
        self._readable.clear()

    def close(self):
        self._epoll.close()

    @property
    def closed(self):
        return self._epoll.closed
    _closed = closed

    def fileno(self):
        return self._epoll.fileno()

    @classmethod
    def fromfd(cls, fd):
        return cls(from_ep=original_epoll_fromfd(fd))

    def modify(self, fd, eventmask):
        self._epoll.modify(fd, eventmask)

    def poll(self, timeout=None, maxevents=-1):
        onr = self._on_readable
        reg = scheduler._register_fd(self._epoll.fileno(), onr, None)
        try:
            self._readable.wait(timeout=timeout)
            return self._epoll.poll(0, maxevents)
        finally:
            scheduler._unregister_fd(self._epoll.fileno(), onr, None, reg)

    def register(self, fd, eventmask=all_epoll_evs):
        self._epoll.register(fd, eventmask)

    def unregister(self, fd):
        self._epoll.unregister(fd)


class green_kqueue(object):
    def __init__(self, from_kq=None):
        self._readable = util.Event()
        self._writable = util.Event()
        if from_kq:
            self._kqueue = from_kq
        else:
            self._kqueue = original_kqueue()

    def _on_readable(self):
        self._readable.set()
        self._readable.clear()

    def _on_writable(self):
        self._writable.set()
        self._writable.clear()

    def close(self):
        self._kqueue.close()

    @property
    def closed(self):
        return self._kqueue.closed
    _closed = closed

    def control(self, events, max_events, timeout=None):
        if not max_events:
            return self._kqueue.control(events, max_events, 0)

        onr, onw = self._on_readable, self._on_writable
        reg = scheduler._register_fd(self._kqueue.fileno(), onr, onw)
        try:
            self._readable.wait(timeout=timeout)
            return self._kqueue.control(events, max_events, 0)
        finally:
            scheduler._unregister_fd(self._kqueue.fileno(), onr, onw, reg)

    def fileno(self):
        return self._kqueue.fileno()

    @classmethod
    def fromfd(cls, fd):
        return cls(from_kq=original_kqueue_fromfd(fd))


patchers = {'select': green_select}

if hasattr(select, "poll"):
    patchers['poll'] = green_poll

if hasattr(select, "epoll"):
    patchers['epoll'] = green_epoll
    original_epoll = select.epoll
    original_epoll_fromfd = select.epoll.fromfd

if hasattr(select, "kqueue"):
    patchers['kqueue'] = green_kqueue
    original_kqueue = select.kqueue
    original_kqueue_fromfd = select.kqueue.fromfd



import collections
import heapq
from Queue import Empty, Full
import time
import weakref

from greenhouse import compat, scheduler


__all__ = ["Event", "Lock", "RLock", "Condition", "Semaphore",
           "BoundedSemaphore", "Timer", "Local", "Thread", "Queue",
           "LifoQueue", "PriorityQueue", "Counter"]


class Event(object):
    """an event for which greenlets can wait

    mirrors the standard library `threading.Event` API
    """
    def __init__(self):
        self._is_set = False
        self._waiters = []

    def is_set(self):
        """indicates whether waiting on the event will block right now

        :returns:
            ``True`` if the event has been :meth:`set` and waiting will not
            block, ``False`` if the event is :meth:`cleared<clear>` and
            :meth:`wait` will block
        """
        return self._is_set
    isSet = is_set

    def set(self):
        """set the event to triggered

        after calling this method, all greenlets waiting on the event will be
        rescheduled, and calling :meth:`wait` will not block until
        :meth:`clear` has been called
        """
        self._is_set = True
        scheduler.state.awoken_from_events.update(self._waiters)
        del self._waiters[:]

    def clear(self):
        """clear the event from being triggered

        after calling this method, waiting on this event will block until the
        :meth:`set` method has been called
        """
        self._is_set = False

    def wait(self, timeout=None):
        """pause the current coroutine until this event is set

        .. note::

            this method will block the current coroutine if :meth:`set` has not
            been called.

        :param timeout:
            the maximum amount of time to block in seconds. the default of
            ``None`` allows indefinite blocking.
        :type timeout: number or None

        :returns:
            ``True`` if a timeout was provided and was hit, otherwise ``False``
        """
        if self._is_set:
            return False

        current = compat.getcurrent()  # the waiting greenlet

        waketime = None if timeout is None else time.time() + timeout
        if timeout is not None:
            scheduler.schedule_at(waketime, current)

        self._waiters.append(current)
        scheduler.state.mainloop.switch()

        if timeout is not None:
            if not scheduler._remove_timer(waketime, current):
                scheduler.state.awoken_from_events.discard(current)
                if current in self._waiters:
                    self._waiters.remove(current)
                return True

        return False


class Lock(object):
    """an object that can only be 'owned' by one greenlet at a time

    mirrors the standard library `threading.Lock` API
    """
    def __init__(self):
        self._locked = False
        self._owner = None
        self._waiters = collections.deque()

    def _is_owned(self):
        return self._owner is compat.getcurrent()

    def locked(self):
        """indicates whether the lock is currently locked

        :returns:
            ``True`` if the lock is locked (and therefore calling
            :meth:`acquire` would block), otherwise ``False``
        """
        return self._locked

    locked_lock = locked

    def acquire(self, blocking=True):
        """lock the lock, blocking until it becomes available

        .. note::

            this method will block the current coroutine if the lock is not
            already owned.

        :param blocking:
            whether to block if the lock is already owned (default ``True``)
        :type blocking: bool

        :returns:
            a `bool` indicating whether the lock was acquired. In the default
            case of ``blocking = True`` this will always be the case, but may
            not be otherwise.
        """
        current = compat.getcurrent()
        if not blocking:
            locked_already = self._locked
            if not locked_already:
                self._locked = True
                self._owner = current
            return not locked_already
        if self._locked:
            self._waiters.append(current)
            scheduler.state.mainloop.switch()
        else:
            self._locked = True
            self._owner = current
        return True

    acquire_lock = acquire

    def release(self):
        """open the lock back up to wake up a waiting greenlet

        :raises:
             `RuntimeError` if the calling greenlet is not the one that had
             :meth:`acquired <acquire>` the lock
        """
        if not self._locked:
            raise RuntimeError("cannot release un-acquired lock")
        if self._waiters:
            waiter = self._waiters.popleft()
            self._locked = True
            self._owner = waiter
            scheduler.state.awoken_from_events.add(waiter)
        else:
            self._locked = False
            self._owner = None

    release_lock = release

    def __enter__(self):
        return self.acquire()

    def __exit__(self, type, value, traceback):
        return self.release()


class RLock(Lock):
    """a lock which may be acquired more than once by the same greenlet

    mirrors the standard library `threading.RLock` API
    """
    def __init__(self):
        super(RLock, self).__init__()
        self._count = 0

    def acquire(self, blocking=True):
        """acquire ownership of the lock

        if the lock is already owned by the calling greenlet, a counter simply
        gets incremented. if it is owned by a different greenlet then it will
        block until the lock becomes available.

        .. note::

            this method will block the current coroutine if the lock is not
            already owned by another coroutine.

        :param blocking:
            whether to block if the lock is owned by a different greenlet
            (default ``True``)
        :type blocking: bool

        :returns:
            a `bool` indicating whether the lock was acquired. In the default
            case of ``blocking = True`` this will always be the case, but may
            not be otherwise.
        """
        current = compat.getcurrent()
        if self._owner is current:
            self._count += 1
            return True
        if self._locked and not blocking:
            return False
        if self._locked:
            self._waiters.append(compat.getcurrent())
            scheduler.state.mainloop.switch()
        else:
            self._locked = True
            self._owner = current
        self._count = 1
        return True

    def release(self):
        """release one ownership of the lock

        if the calling greenlet has :meth:`acquired <acquire>` the lock more
        than once this will simply decrement the counter. if this is a final
        release then a waiting greenlet is awoken

        :raises:
            `RuntimeError` if the calling greenlet is not the lock's owner
        """
        if not self._locked or self._owner is not compat.getcurrent():
            raise RuntimeError("cannot release un-acquired lock")
        self._count -= 1
        if self._count == 0:
            self._owner = None
            if self._waiters:
                waiter = self._waiters.popleft()
                self._locked = True
                self._owner = waiter
                scheduler.state.awoken_from_events.add(waiter)
            else:
                self._locked = False
                self._owner = None


class Condition(object):
    """a synchronization object capable of waking all or one of its waiters

    mirrors the standard library `threading.Condition` API

    :param lock: the lock object wrapped by the condition
    :type lock: :class:`Lock` or :class:`RLock`
    """
    def __init__(self, lock=None):
        if lock is None:
            lock = RLock()
        self._lock = lock
        self._waiters = collections.deque()
        self.acquire = lock.acquire
        self.release = lock.release
        self.__enter__ = lock.__enter__
        self.__exit__ = lock.__exit__
        if hasattr(lock, '_is_owned'):
            self._is_owned = lock._is_owned

    def _is_owned(self):
        owned = not self._lock.acquire(False)
        if not owned:
            self._lock.release()
        return owned

    def wait(self, timeout=None):
        """wait to be woken up by the condition

        .. note::

            this method will block the current coroutine until a :meth:`notify`
            wakes it back up.

        :raises:
            `RuntimeError` if the underlying lock hasn't been
            :meth:`acquired <Lock.acquire>`
        """
        if not self._is_owned():
            raise RuntimeError("cannot wait on un-acquired lock")

        current = compat.getcurrent()

        waketime = None if timeout is None else time.time() + timeout
        if timeout is not None:
            scheduler.schedule_at(waketime, current)
        self._waiters.append((current, waketime))

        self._lock.release()
        scheduler.state.mainloop.switch()
        self._lock.acquire()

        if timeout is not None:
            timedout = not scheduler._remove_timer(waketime, current)
            if timedout:
                self._waiters.remove((current, waketime))
            return timedout

        return False

    def notify(self, num=1):
        """wake one or more waiting greenlets

        :param num: the number of waiters to wake (default 1)
        :type num: int

        :raises:
            `RuntimeError` if the underlying lock hasn't been
            :meth:`acquired <Lock.acquire>`
        """
        if not self._is_owned():
            raise RuntimeError("cannot wait on un-acquired lock")
        for i in xrange(min(num, len(self._waiters))):
            scheduler.state.awoken_from_events.add(self._waiters.popleft()[0])

    def notify_all(self):
        """wake all waiting greenlets

        :raises:
            `RuntimeError` if the underlying lock hasn't been
            :meth:`acquired <Lock.acquire>`
        """
        if not self._is_owned():
            raise RuntimeError("cannot wait on un-acquired lock")
        scheduler.state.awoken_from_events.update(x[0] for x in self._waiters)
        self._waiters.clear()
    notifyAll = notify_all


class Semaphore(object):
    """a synchronization object with a counter that blocks when it reaches 0

    mirrors the api of `threading.Semaphore`

    :param value:
        the starting value of the counter
    :type value: int
    """
    def __init__(self, value=1):
        assert value >= 0, "semaphore value cannot be negative"
        self._value = value
        self._waiters = collections.deque()

    def acquire(self, blocking=True):
        """decrement the counter, waiting if it is already at 0

        .. note::

            if the counter is already at 0, this method will block the current
            coroutine until a :meth:`release` increments it again.

        :param blocking:
            whether or not to block if the counter is already at 0 (default
            ``True``)
        :type blocking: bool

        :returns:
            a bool, indicating whether the count was decremented (this can only
            be ``False`` if ``blocking`` was ``False`` -- otherwise it would
            have blocked until it could decrement the counter)
        """
        if self._value:
            self._value -= 1
            return True
        if not blocking:
            return False
        self._waiters.append(compat.getcurrent())
        scheduler.state.mainloop.switch()
        return True

    def release(self):
        "increment the counter, waking up a waiter if there was any"
        if self._waiters:
            scheduler.state.awoken_from_events.add(self._waiters.popleft())
        else:
            self._value += 1

    def __enter__(self):
        return self.acquire()

    def __exit__(self, type, value, traceback):
        return self.release()


class BoundedSemaphore(Semaphore):
    """a semaphore with an upper limit to the counter

    mirrors the api of `threading.BoundedSemaphore`

    :param value:
        the starting and maximum value of the counter
    :type value: int
    """
    def __init__(self, value=1):
        super(BoundedSemaphore, self).__init__(value)
        self._initial_value = value

    def release(self):
        """increment the counter, waking up a waiter if there was any

        .. note::

            if the counter is already at 0, this method will block the current
            coroutine until a :meth:`release` increments it again.

        :raises: `ValueError` if this increment would have crossed the maximum
        """
        if self._value >= self._initial_value:
            raise ValueError("BoundedSemaphore released too many times")
        return super(BoundedSemaphore, self).release()
    release.__doc__ = Semaphore.release.__doc__


class Local(object):
    """an object that holds greenlet-local data

    mirrors the standard library `threading.local` API

    to use, simply create an instance with no arguments, and any attribute gets
    and sets will be specific to that greenlet
    """
    _main_standin = type('', (), {})()

    def __init__(self):
        object.__setattr__(self, "_local_data", weakref.WeakKeyDictionary())

    def __getattr__(self, name):
        current = compat.getcurrent()
        if current is compat.main_greenlet:
            current = self._main_standin
        local = self._local_data.setdefault(current, {})
        if name not in local:
            raise AttributeError("Local object has no attribute %s" % name)
        return local[name]

    def __setattr__(self, name, value):
        current = compat.getcurrent()
        if current is compat.main_greenlet:
            current = self._main_standin
        self._local_data.setdefault(current, {})[name] = value


class Thread(object):
    """a standin class for threads, but powered by greenlets

    this class is useful for :mod:`monkey-patching <greenhouse.emulation>`
    the standard library, but for code written explicitly for greenhouse, the
    functions in :mod:`greenhouse.scheduler` are a better way to go

    mirrors the standard library `threading.Thread` API

    :param group:
        this argument does nothing and must be ``None`` (for compatibility with
        the standard library `threading.Thread`)
    :type group: None
    :param target: the function to run in the greenlet
    :type target: function
    :param name: the thread name (defaults to a generated name)
    :type name: str
    :param args: positional arguments to pass in to the `target` function
    :type args: tuple
    :param kwargs: keyword arguments to pass to the `target` function
    :type kwargs: dict
    :param verbose: here for compatibility, it is actually ignored
    :type verbose: bool
    """
    def __init__(
            self,
            group=None,
            target=None,
            name=None,
            args=(),
            kwargs=None,
            verbose=None):
        assert group is None, "group argument must be None for now"  # [sic]

        self._target = target
        self.name = str(name or self._newname())
        self._args = args or ()
        self._kwargs = kwargs or {}
        self._started = False
        self._finished = Event()
        self._glet = None
        self._ident = None
        # verbose is ignored

    def __repr__(self):
        status = "initial"
        if self._started:
            status = "started"
        if self._finished.is_set():
            status = "stopped"
        return "<%s (%s, %s)>" % (type(self).__name__, self.name, status)

    def _activate(self):
        self._started = True
        type(self)._active[self._glet] = self
        type(self)._active_by_id[self.ident] = self

    def _deactivate(self):
        self._finished.set()
        type(self)._active.pop(self._glet)
        type(self)._active_by_id.pop(self.ident)

    def start(self):
        """schedule to start the greenlet that runs this thread's function

        :raises: `RuntimeError` if the thread has already been started
        """
        if self._started:
            raise RuntimeError("thread already started")

        def run():
            try:
                self.run(*self._args, **self._kwargs)
            except SystemExit:
                # only shut down the thread, not the whole process
                pass
            finally:
                self._deactivate()

        self._glet = scheduler.greenlet(run)
        self._ident = id(self._glet)
        scheduler.schedule(self._glet)
        self._activate()

    def run(self, *args, **kwargs):
        """override this method to define the thread's behavior

        this is an alternative way to define the thread's function than passing
        `target` to the constructor
        """
        self._target(*args, **kwargs)

    def join(self, timeout=None):
        """block until this thread terminates

        .. note::

            this method can block the calling coroutine if the thread has not
            yet completed.

        :param timeout:
            the maximum time to wait. with the default of ``None``, waits
            indefinitely
        :type timeout: int, float or None

        :raises:
            `RuntimeError` if called inside the thread, or it has not yet been
            started
        """
        if not self._started:
            raise RuntimeError("cannot join thread before it is started")
        if compat.getcurrent() is self._glet:
            raise RuntimeError("cannot join current thread")
        self._finished.wait(timeout)

    @property
    def ident(self):
        "unique identifier for this thread"
        return self._ident

    def is_alive(self):
        """indicates whether the thread is currently running

        :returns:
            ``True`` if :meth:`start` has already been called but the function
            hasn't yet finished or been killed, ``False`` if it isn't currently
            running for any reason.
        """
        return self._started and not self._finished.is_set()
    isAlive = is_alive

    def is_daemon(self):
        """whether the thread is in daemonized mode

        :returns:
            ``True``. this is here for compatibility with `threading.Thread`,
            greenhouse-based threads always operate like daemonized threads.
        """
        return True
    isDaemon = is_daemon

    def set_daemon(self, daemonic):
        """here for compatibility with `threading.Thread` (this doesn't work).

        :param daemonic: whether attempting to turn daemon mode on or off
        :type daemonic: bool
        """
        pass
    setDaemon = set_daemon

    daemon = property(is_daemon, set_daemon,
                      doc="whether the thread is set as a " +
                          "daemon thread (unsupported)")

    def get_name(self):
        """the thread's name as passed in the constructor or :meth:`set_name`,
        or failing those the automatically generated name

        :returns: the thread's `str` name
        """
        return self.name
    getName = get_name

    def set_name(self, name):
        """overwrite the thread's name

        :param name: the name to set
        :type name: str
        """
        self.name = name
    setName = set_name

    _counter = 0

    @classmethod
    def _newname(cls, template="Thread-%d"):
        cls._counter += 1
        return template % cls._counter

    _active = {}
    _active_by_id = {}


class Timer(Thread):
    """creates a greenlet from a function and schedules it to run later

    mirrors the standard library `threading.Timer` API

    :param interval: how far in the future to defer the function in seconds
    :type interval: int or float
    :param func: the function to run later
    :type func: function
    :param args: positional arguments to pass to the function
    :type args: tuple
    :param kwargs: keyword arguments to pass to the function
    :type kwargs: dict
    """
    def __init__(self, interval, function, args=[], kwargs={}):
        super(Timer, self).__init__()
        self.interval = interval
        self.function = function
        self.args = args
        self.kwargs = kwargs
        self.finished = Event()

    def cancel(self):
        """attempt to prevent the timer from ever running its function

        :returns:
            ``True`` if it was successful, ``False`` if the timer had already
            run or been cancelled
        """
        done = self.finished.is_set()
        self.finished.set()
        return not done

    def run(self):
        self.finished.wait(self.interval)
        if not self.finished.is_set():
            self.function(*self.args, **self.kwargs)
        self.finished.set()

    @classmethod
    def wrap(cls, secs, args=(), kwargs=None):
        """a classmethod decorator to immediately turn a function into a timer

        .. note::

            you won't find this on `threading.Timer`, it is an extension to
            that API

        this is a function *returning a decorator*, so it is used like so:

        >>> @Timer.wrap(5, args=("world",))
        >>> def say_hi_timer(name):
        ...     print "hello, %s" % name

        :param secs:
            how far in the future to defer running the function in seconds
        :type secs: int or float
        :param args: positional arguments to pass to the function
        :type args: tuple
        :param kwargs: keyword arguments to pass to the function
        :type kwargs: dict

        :returns:
            a decorator function which produces an unstarted :class:`Timer`
        """
        def decorator(func):
            return cls(secs, func, args, kwargs)
        return decorator


class Queue(object):
    """a producer-consumer queue

    :param maxsize:
        optional limit to the amount of queued data, after which :meth:`put`
        can block. the default of 0 turns off the limit, so :meth:`put` will
        never block
    :type maxsize: int

    mirrors the standard library `Queue.Queue` API
    """
    _data_type = collections.deque

    def __init__(self, maxsize=0):
        self._maxsize = maxsize
        self._waiters = collections.deque()
        self._data = self._data_type()
        self._open_tasks = 0
        self._jobs_done = Event()
        self._jobs_done.clear()

    def empty(self):
        """indicate whether there is any queued data

        :returns: ``True`` if there is data in the queue, otherwise ``False``
        """
        return not self._data

    def full(self):
        """indicate whether the queue has `maxsize` data queued

        :returns:
            ``True`` if the queue's data has reached `maxsize`, otherwise
            ``False``
        """
        return self._maxsize > 0 and len(self._data) == self._maxsize

    def _get(self):
        return self._data.popleft()

    def get(self, block=True, timeout=None):
        """get an item out of the queue

        .. note::

            if `block` is ``True`` (the default) and the queue is
            :meth`empty`, this method will block the current coroutine until
            something has been :meth:`put`.

        :param block:
            whether to block if there is no data yet available (default
            ``True``)
        :type block: bool
        :param timeout:
            the maximum time in seconds to block waiting for data. with the
            default of ``None``, can wait indefinitely. this is unused if
            `block` is ``False``.
        :type timeout: int, float or None

        :raises:
            :class:`Empty` if there is no data in the queue and block is
            ``False``, or `timeout` expires

        :returns: something that was previously :meth:`put` in the queue
        """
        if not self._data:
            if not block:
                raise Empty()

            current = compat.getcurrent()

            waketime = None if timeout is None else time.time() + timeout
            if timeout is not None:
                scheduler.schedule_at(waketime, current)
            self._waiters.append((current, waketime))

            scheduler.state.mainloop.switch()

            if timeout is not None:
                if not scheduler._remove_timer(waketime, current):
                    self._waiters.remove((current, waketime))
                    raise Empty()

        if self.full() and self._waiters:
            scheduler.schedule(self._waiters.popleft()[0])

        return self._get()

    def get_nowait(self):
        """get an item out of the queue without ever blocking

        this call is equivalent to ``get(block=False)``

        :returns: something that was previously :meth:`put` in the queue
        """
        return self.get(block=False)

    def _put(self, item):
        self._data.append(item)

    def put(self, item, block=True, timeout=None):
        """put an item into the queue

        .. note::

            if the queue was created with a `maxsize` and it is currently
            :meth:`full`, this method will block the calling coroutine until
            another coroutine :meth:`get`\ s an item.

        :param item: the object to put into the queue, can be any type
        :param block:
            whether to block if the queue is already :meth:`full` (default
            ``True``)
        :type block: bool
        :param timeout:
            the maximum time in seconds to block waiting. with the default of
            ``None``, it can wait indefinitely. this is unused if `block` is
            ``False``.
        :type timeout: int, float or None

        :raises:
            :class:`Full` if the queue is :meth:`full` and `block` is
            ``False``, or if `timeout` expires.
        """
        if self.full():
            if not block:
                raise Full()

            current = compat.getcurrent()

            waketime = None if timeout is None else time.time() + timeout
            if timeout is not None:
                scheduler.schedule_at(waketime, current)
            self._waiters.append((current, waketime))

            scheduler.state.mainloop.switch()

            if timeout is not None:
                if not scheduler._remove_timer(waketime, current):
                    self._waiters.remove((current, waketime))
                    raise Full()

        if self._waiters and not self.full():
            scheduler.schedule(self._waiters.popleft()[0])

        if not self._open_tasks:
            self._jobs_done.clear()
        self._open_tasks += 1

        self._put(item)

    def put_nowait(self, item):
        """put an item into the queue without any chance of blocking

        this call is equivalent to ``put(block=False)``

        ;param item: the object to put into the queue, can be any type
        """
        return self.put(item, block=False)

    def qsize(self):
        """the number of queued pieces of data

        :returns: int
        """
        return len(self._data)

    def task_done(self):
        """mark that a "job" (corresponding to a :meth:`put` or
        :meth:`put_nowait` call) is finished

        the :meth:`join` method won't complete until the number of
        :meth:`task_done` calls equals the number of :meth:`put` calls
        """
        if not self._open_tasks:
            raise ValueError("task_done() called too many times")
        self._open_tasks -= 1
        if not self._open_tasks:
            self._jobs_done.set()

    def join(self, timeout=None):
        """wait for task completions

        .. note::

            if :meth:`task_done` has not been called for every :meth:`put`
            call, this method will block until it has.

        the queue is still usable after a :meth:`join` call. a return from
        :meth:`join` simply indicates that the queue has no jobs `currently`
        pending.

        :param timeout:
            the maximum amount of time to wait in seconds. the default of
            ``None`` allows indefinite waiting.
        :type timeout: int, float or None

        :returns:
            ``True`` if `timeout` was provided and expired, otherwise ``False``
        """
        return self._jobs_done.wait(timeout)


class LifoQueue(Queue):
    """a producer-consumer queue that produces items in LIFO order

    :param maxsize:
        optional limit to the amount of queued data, after which :meth:`put`
        can block. the default of 0 turns off the limit, so :meth:`put` will
        never block
    :type maxsize: int

    mirrors the standard library `Queue.LifoQueue` API
    """
    _data_type = list

    def _get(self):
        return self._data.pop()

    def _put(self, item):
        self._data.append(item)


class PriorityQueue(Queue):
    """a producer-consumer queue that produces items in prioritized order

    :param maxsize:
        optional limit to the amount of queued data, after which :meth:`put`
        can block. the default of 0 turns off the limit, so :meth:`put` will
        never block
    :type maxsize: int

    mirrors the standard library `Queue.PriorityQueue` API
    """
    _data_type = list

    def _get(self):
        return heapq.heappop(self._data)

    def _put(self, item):
        heapq.heappush(self._data, item)


class Counter(object):
    "a counter object that can block"
    def __init__(self, initial=0):
        self._count = initial
        self._waiters = {}

    def increment(self):
        "increment the counter, and wake anyone waiting for the new value"
        self._count += 1
        waiters = self._waiters.pop(self._count, [])
        if waiters:
            scheduler.state.awoken_from_events.update(waiters)

    __enter__ = increment

    def decrement(self):
        "decrement the counter and wake anyone waiting for the new value"
        self._count -= 1
        waiters = self._waiters.pop(self._count, [])
        if waiters:
            scheduler.state.awoken_from_events.update(waiters)

    def __exit__(self, etype, exc, tb):
        self.decrement()

    @property
    def count(self):
        "the current integer value of the counter"
        return self._count

    def wait(self, until=0):
        """wait until the count has reached a particular number

        .. note:: this method can block the current greenlet

        :param until:
            the number to wait for the count to get down (or up) to. default 0
        :type until: int
        """
        if self._count != until:
            self._waiters.setdefault(until, []).append(compat.getcurrent())
            scheduler.state.mainloop.switch()



from __future__ import absolute_import

import psycopg2
from psycopg2 import extensions

from ..io import descriptor


__all__ = ["wait_callback"]


def wait_callback(connection):
    """callback function suitable for ``psycopg2.set_wait_callback``

    pass this function to ``psycopg2.extensions.set_wait_callack`` to force any
    blocking operations from psycopg2 to only block the current coroutine,
    rather than the entire thread or process

    to undo the change and return to normal blocking operation, use
    `psycopg2.extensions.set_wait_callback(None)``
    """
    while 1:
        state = connection.poll()

        if state == extensions.POLL_OK:
            break
        elif state == extensions.POLL_READ:
            descriptor.wait_fds([(connection.fileno(), 1)])
        elif state == extensions.POLL_WRITE:
            descriptor.wait_fds([(connection.fileno(), 2)])
        else:
            raise psycopg2.OperationalError("Bad poll result: %r" % state)



from __future__ import absolute_import

import logging
import sys

from greenhouse.compat import *
from greenhouse.scheduler import *
from greenhouse.util import *
from greenhouse.pool import *

from greenhouse.io import *
from greenhouse.backdoor import *
from greenhouse.emulation import *


VERSION = (2, 1, 13, '')

__version__ = ".".join(filter(None, (str(x) for x in VERSION)))

reset_poller()


# prime the pump. if there is a traceback before the mainloop greenlet
# has a chance to get into its 'try' block, the mainloop will die of that
# traceback and it will wind up being raised in the main greenlet
@schedule
def f():
    pass
pause()
del f


def configure_logging(filename=None, filemode=None, fmt=None,
                      level=logging.INFO, stream=None, handler=None):
    if handler is None:
        if filename is None:
            handler = logging.StreamHandler(stream or sys.stderr)
        else:
            handler = logging.FileHandler(filename, filemode or 'a')

    if fmt is None:
        fmt = "[%(asctime)s] %(name)s/%(levelname)s | %(message)s"
    handler.setFormatter(logging.Formatter(fmt))

    log = logging.getLogger("greenhouse")
    log.setLevel(level)
    log.addHandler(handler)



from __future__ import absolute_import

import collections
import errno
import operator
import select
import sys

_original_select = select.select


class Poll(object):
    "a greenhouse poller using the poll system call"
    INMASK = getattr(select, 'POLLIN', 0)
    OUTMASK = getattr(select, 'POLLOUT', 0)
    ERRMASK = getattr(select, 'POLLERR', 0) | getattr(select, "POLLHUP", 0)

    _POLLER = getattr(select, "poll", None)

    def __init__(self):
        self._poller = self._POLLER()
        self._registry = collections.defaultdict(dict)
        self._counter = 0

    def register(self, fd, eventmask=None):
        # integer file descriptor
        fd = fd if isinstance(fd, int) else fd.fileno()

        # mask nothing by default
        if eventmask is None:
            eventmask = self.INMASK | self.OUTMASK | self.ERRMASK

        # get the current registrations dictionary
        registrations = self._registry[fd]
        registered = reduce(
            operator.or_, registrations.itervalues(), 0)

        # update registrations in the OS poller
        self._update_registration(fd, registered, registered | eventmask)

        # store the registration
        self._counter += 1
        registrations[self._counter] = eventmask

        return self._counter

    def unregister(self, fd, counter):
        # integer file descriptor
        fd = fd if isinstance(fd, int) else fd.fileno()

        registrations = self._registry[fd]

        # allow for extra noop calls
        if counter not in registrations:
            self._registry.pop(fd)
            return

        mask = registrations.pop(counter)
        the_rest = reduce(operator.or_, registrations.itervalues(), 0)

        # update the OS poller's registration
        self._update_registration(fd, the_rest | mask, the_rest)

        if not registrations:
            self._registry.pop(fd)

    def poll(self, timeout):
        if timeout is not None:
            timeout *= 1000
        return self._poller.poll(timeout)

    def _update_registration(self, fd, from_mask, to_mask):
        if from_mask != to_mask:
            if from_mask and to_mask:
                self._poller.modify(fd, to_mask)
            elif from_mask:
                self._poller.unregister(fd)
            elif to_mask:
                self._poller.register(fd, to_mask)

    def supports(self, fd):
        if not isinstance(fd, int):
            fd = fd.fileno()

        try:
            self._poller.register(fd)
        except EnvironmentError:
            return False

        self._poller.unregister(fd)
        return True


class Epoll(Poll):
    "a greenhouse poller utilizing the 2.6+ stdlib's epoll support"
    INMASK = getattr(select, 'EPOLLIN', 0)
    OUTMASK = getattr(select, 'EPOLLOUT', 0)
    ERRMASK = getattr(select, 'EPOLLERR', 0) | getattr(select, "EPOLLHUP", 0)

    _POLLER = getattr(select, "epoll", None)

    def poll(self, timeout):
        if timeout is None:
            timeout = -1
        return self._poller.poll(timeout)


class KQueue(Poll):
    "a greenhouse poller using the 2.6+ stdlib's kqueue support"
    INMASK = 1
    OUTMASK = 2
    ERRMASK = 0

    _POLLER = getattr(select, "kqueue", None)

    _mask_map = {
        getattr(select, "KQ_FILTER_READ", 0): INMASK,
        getattr(select, "KQ_FILTER_WRITE", 0): OUTMASK,
    }

    def poll(self, timeout):
        evs = self._poller.control(None, 2 * len(self._registry), timeout)
        return [(ev.ident, self._mask_map[ev.filter]) for ev in evs]

    def _update_registration(self, fd, from_mask, to_mask):
        if from_mask == to_mask:
            return

        xor = from_mask ^ to_mask
        to_add = to_mask & xor
        to_drop = from_mask & xor
        assert not to_add & to_drop  # simple sanity

        events = []
        if to_add & self.INMASK:
            events.append(select.kevent(
                fd, select.KQ_FILTER_READ, select.KQ_EV_ADD))
        elif to_drop & self.INMASK:
            events.append(select.kevent(
                fd, select.KQ_FILTER_READ, select.KQ_EV_DELETE))
        if to_add & self.OUTMASK:
            events.append(select.kevent(
                fd, select.KQ_FILTER_WRITE, select.KQ_EV_ADD))
        elif to_drop & self.OUTMASK:
            events.append(select.kevent(
                fd, select.KQ_FILTER_WRITE, select.KQ_EV_DELETE))

        if events:
            if sys.platform == 'darwin':
                # busted OS X kqueue only accepts 1 kevent at a time
                for event in events:
                    self._poller.control([event], 0)
            else:
                self._poller.control(events, 0)


class Select(object):
    "a greenhouse poller using the select system call"
    INMASK = 1
    OUTMASK = 2
    ERRMASK = 4

    def __init__(self):
        self._registry = collections.defaultdict(dict)
        self._currentmasks = {}
        self._counter = 0

    def register(self, fd, eventmask=None):
        # integer file descriptor
        fd = fd if isinstance(fd, int) else fd.fileno()

        # mask nothing by default
        if eventmask is None:
            eventmask = self.INMASK | self.OUTMASK | self.ERRMASK

        # store the registration
        self._counter += 1
        self._registry[fd][self._counter] = eventmask

        # update the full mask
        self._currentmasks[fd] = self._currentmasks.get(fd, 0) | eventmask

        return self._counter

    def unregister(self, fd, counter):
        # integer file descriptor
        fd = fd if isinstance(fd, int) else fd.fileno()

        # just drop it from the registrations dict
        self._registry.get(fd, {}).pop(counter, None)

        # rewrite the full mask
        newmask = reduce(
            operator.or_, self._registry[fd].itervalues(), 0)
        if newmask:
            self._currentmasks[fd] = newmask
        else:
            self._currentmasks.pop(fd, 0)

        if not self._registry[fd]:
            self._registry.pop(fd)

    def poll(self, timeout):
        rlist, wlist, xlist = [], [], []
        for fd, eventmask in self._currentmasks.iteritems():
            if eventmask & self.INMASK:
                rlist.append(fd)
            if eventmask & self.OUTMASK:
                wlist.append(fd)
            if eventmask & self.ERRMASK:
                xlist.append(fd)
        rlist, wlist, xlist = _original_select(rlist, wlist, xlist, timeout)
        events = collections.defaultdict(int)
        for fd in rlist:
            events[fd] |= self.INMASK
        for fd in wlist:
            events[fd] |= self.OUTMASK
        for fd in xlist:
            events[fd] |= self.ERRMASK
        return events.items()

    def supports(self, fd):
        # it might lie though
        return True


def best():
    if hasattr(select, 'epoll'):
        return Epoll()

    if hasattr(select, 'kqueue'):
        return KQueue()

    if hasattr(select, 'poll'):
        return Poll()

    return Select()



from __future__ import absolute_import

import time

from ..io import descriptor
import zmq.backend


__all__ = ["wait_socks"]


def wait_socks(sock_events, inmask=1, outmask=2, timeout=None):
    """wait on a combination of zeromq sockets, normal sockets, and fds

    .. note:: this method can block

        it will return once there is relevant activity on any of the
        descriptors or sockets, or the timeout expires

    :param sock_events:
        two-tuples, the first item is either a zeromq socket, a socket, or a
        file descriptor, and the second item is a mask made up of the inmask
        and/or the outmask bitwise-ORd together
    :type sock_events: list
    :param inmask: the mask to use for readable events (default 1)
    :type inmask: int
    :param outmask: the mask to use for writable events (default 2)
    :type outmask: int
    :param timeout: the maximum time to block before raising an exception
    :type timeout: int, float or None

    :returns:
        a list of two-tuples, each has one of the first elements from
        ``sock_events``, the second element is the event mask of the activity
        that was detected (made up on inmask and/or outmask bitwise-ORd
        together)
    """
    results = []
    for sock, mask in sock_events:
        if isinstance(sock, zmq.backend.Socket):
            mask = _check_events(sock, mask, inmask, outmask)
            if mask:
                results.append((sock, mask))
    if results:
        return results

    fd_map = {}
    fd_events = []
    for sock, mask in sock_events:
        if isinstance(sock, zmq.backend.Socket):
            fd = sock.getsockopt(zmq.FD)
        elif isinstance(sock, int):
            fd = sock
        else:
            fd = sock.fileno()

        fd_map[fd] = sock
        fd_events.append((fd, mask))

    while 1:
        started = time.time()
        active = descriptor.wait_fds(fd_events, inmask, outmask, timeout)
        if not active:
            # timed out
            return []

        results = []
        for fd, mask in active:
            sock = fd_map[fd]
            if isinstance(sock, zmq.backend.Socket):
                mask = _check_events(sock, mask, inmask, outmask)
                if not mask:
                    continue
            results.append((sock, mask))

        if results:
            return results

        timeout -= time.time() - started


def _check_events(sock, mask, inmask=1, outmask=2):
    evs = sock.getsockopt(zmq.EVENTS)
    result = 0
    if evs & zmq.POLLIN and mask & inmask:
        result |= inmask
    if evs & zmq.POLLOUT and mask & outmask:
        result |= outmask
    return result



from __future__ import absolute_import, with_statement

import functools
import socket

from ..ext import dns


@functools.wraps(socket.getaddrinfo)
def getaddrinfo(host, port, family=0, socktype=0, proto=0, flags=0):
    socktype = socktype or socket.SOCK_STREAM
    try:
        addrs = dns.resolve(host)
    except dns.resolver.NXDOMAIN:
        raise socket.gaierror(-2, "Name or service not known")
    return [(socket.AF_INET, socktype, proto, '', (addr, port))
            for addr in addrs]


@functools.wraps(socket.gethostbyname)
def gethostbyname(hostname):
    return dns.resolve(hostname)[0]


@functools.wraps(socket.gethostbyname_ex)
def gethostbyname_ex(hostname):
    dns.build_resolver()
    results = dns.resolve(hostname)
    try:
        canon = dns.resolver_obj.query(hostname, dns.rdatatype.CNAME)
        aliases = [hostname]
    except dns.resolver.NoAnswer:
        canon = hostname
        aliases = []

    return canon, aliases, results


@functools.wraps(socket.getnameinfo)
def getnameinfo(address, flags):
    dns.build_resolver()
    try:
        host, port = address
    except (ValueError, TypeError):
        if not isinstance(address, tuple):
            del address
            raise TypeError('getnameinfo() argument must be a tuple')
        else:
            raise socket.gaierror(
                socket.EAI_NONAME, "Name or service not known")

    if (flags & socket.NI_NAMEREQD) and (flags & socket.NI_NUMERICHOST):
        raise socket.gaierror(
            socket.EAI_NONAME, "Name or service not known")

    if dns.is_ipv4(host):
        try:
            name = dns.reversename.from_address(host)

            results = dns.resolver_obj.query(name, dns.rdatatype.PTR)
            if len(results) > 1:
                raise socket.error(
                    "sockaddr resolved to multiple addresses")

            host = results[0].target.to_text(omit_final_dot=True)
        except dns.exception.Timeout, exc:
            if flags & socket.NI_NAMEREQD:
                raise socket.gaierror(socket.EAI_AGAIN, 'Lookup timed out')
        except dns.resolver.NXDOMAIN:
            return (host, str(port))
        except dns.exception.DNSException, exc:
            if flags & socket.NI_NAMEREQD:
                raise socket.gaierror(
                    socket.EAI_NONAME, "Name or service not known")

    else:
        try:
            ips = dns.resolve(host)

            if len(ips) > 1:
                raise socket.error('sockaddr resolved to multiple addresses')

            if flags & socket.NI_NUMERICHOST:
                host = ips[0].to_text()
        except dns.exception.Timeout, exc:
            if flags & socket.NI_NAMEREQD:
                raise socket.gaierror(socket.EAI_AGAIN, 'Lookup timed out')
        except dns.exception.DNSException, exc:
            if flags & socket.NI_NAMEREQD:
                raise socket.gaierror(
                    socket.EAI_NONAME, "Name or service not known")

    if flags & socket.NI_NUMERICSERV:
        port = str(port)
    else:
        port = socket.getservbyport(
            port, (flags & socket.NI_DGRAM) and 'udp' or 'tcp')

    return host, port


socket_patchers = {
    'getaddrinfo': getaddrinfo,
    'gethostbyname': gethostbyname,
    'gethostbyname_ex': gethostbyname_ex,
    'getnameinfo': getnameinfo,
}



from __future__ import absolute_import

import functools
import signal

from greenhouse import scheduler


original_signal = signal.signal


def generic_handler(delegate, *args, **kwargs):
    scheduler.schedule(delegate, args=args, kwargs=kwargs)


def green_signal(sig, action):
    original_signal(sig, functools.partial(generic_handler, action))


patchers = {
    'signal': green_signal,
}



from __future__ import absolute_import

import errno
import fcntl
import functools
import os
import sys

from .. import io, scheduler


OS_TIMEOUT = 0.01

original_fcntl = fcntl.fcntl


def nonblocking_fd(fd):
    flags = original_fcntl(fd, fcntl.F_GETFL)
    if not flags & os.O_NONBLOCK:
        original_fcntl(fd, fcntl.F_SETFL, flags | os.O_NONBLOCK)
    return flags


original_os_read = os.read
original_os_write = os.write
original_os_waitpid = os.waitpid
original_os_wait = os.wait
original_os_wait3 = os.wait3
original_os_wait4 = os.wait4
original_os_popen = os.popen
original_os_popen2 = os.popen2
original_os_popen3 = os.popen3
original_os_popen4 = os.popen4
original_os_system = os.system
original_os_spawnl = os.spawnl
original_os_spawnle = os.spawnle
original_os_spawnlp = os.spawnlp
original_os_spawnlpe = os.spawnlpe
original_os_spawnv = os.spawnv
original_os_spawnve = os.spawnve
original_os_spawnvp = os.spawnvp
original_os_spawnvpe = os.spawnvpe


def green_read(fd, buffsize):
    flags = nonblocking_fd(fd)
    if flags & os.O_NONBLOCK:
        return original_os_read(fd, buffsize)

    try:
        while 1:
            try:
                return original_os_read(fd, buffsize)
            except EnvironmentError, exc:
                if exc.args[0] != errno.EAGAIN:
                    raise
                io.wait_fds([(fd, 1)])
    finally:
        if not flags & os.O_NONBLOCK:
            original_fcntl(fd, fcntl.F_SETFL, flags)


def green_write(fd, data):
    flags = nonblocking_fd(fd)
    if flags & os.O_NONBLOCK:
        return original_os_write(fd, data)

    try:
        while 1:
            try:
                return original_os_write(fd, data)
            except EnvironmentError, exc:
                if exc.args[0] != errno.EAGAIN:
                    raise
                io.wait_fds([(fd, 2)])
    finally:
        if not flags & os.O_NONBLOCK:
            original_fcntl(fd, fcntl.F_SETFL, flags)


def blocking_read(fd, buffsize):
    nonblocking_fd(fd)
    while 1:
        try:
            return original_os_read(fd, buffsize)
        except EnvironmentError, exc:
            if exc.args[0] != errno.EAGAIN:
                raise
            io.wait_fds([(fd, 1)])


def blocking_write(fd, data):
    nonblocking_fd(fd)
    while 1:
        try:
            return original_os_write(fd, data)
        except EnvironmentError, exc:
            if exc.args[0] != errno.EAGAIN:
                raise
            io.wait_fds([(fd, 2)])


def polling_green_version(func, retry_test, opt_arg_num, arg_count, timeout):
    @functools.wraps(func)
    def green_version(*args):
        if len(args) < arg_count:
            raise TypeError("%s takes %d arguments" % (
                func.__name__, arg_count))

        options = args[opt_arg_num]
        if options & os.WNOHANG:
            return func(*args)

        args = list(args)
        args[opt_arg_num] = options | os.WNOHANG
        while 1:
            result = func(*args)
            if not retry_test(result):
                return result
            scheduler.pause_for(timeout)

    return green_version

green_waitpid = polling_green_version(
    original_os_waitpid, lambda x: not (x[0] or x[1]), 1, 2, OS_TIMEOUT)
green_wait3 = polling_green_version(
    original_os_wait3, lambda x: not (x[0] or x[1]), 0, 1, OS_TIMEOUT)
green_wait4 = polling_green_version(
    original_os_wait4, lambda x: not (x[0] or x[1]), 1, 2, OS_TIMEOUT)


@functools.wraps(original_os_wait)
def green_wait():
    return green_waitpid(0, 0)


class green_popen_pipe(io.File):
    @classmethod
    def fromfd(cls, fd, mode, pid):
        fp = super(green_popen_pipe, cls).fromfd(fd, mode)
        fp._pid = pid
        return fp

    def close(self):
        super(green_popen_pipe, self).close()
        return green_waitpid(self._pid, 0)[1] or None


@functools.wraps(original_os_popen)
def green_popen(cmd, mode='r', bufsize=-1):
    if 'r' in mode:
        pipe = 'stdout'
    if 'w' in mode:
        pipe = 'stderr'
    kwarg = {pipe: green_subprocess.PIPE}

    proc = green_subprocess.Popen(cmd, shell=1, bufsize=bufsize, **kwarg)
    return green_popen_pipe.fromfd(
        getattr(proc, pipe).fileno(), mode, proc.pid)


@functools.wraps(original_os_popen2)
def green_popen2(cmd, mode='r', bufsize=-1):
    proc = green_subprocess.Popen(cmd, shell=1, bufsize=bufsize,
                                  stdin=green_subprocess.PIPE,
                                  stdout=green_subprocess.PIPE)
    return proc.stdin, proc.stdout


@functools.wraps(original_os_popen3)
def green_popen3(cmd, mode='r', bufsize=-1):
    proc = green_subprocess.Popen(cmd, shell=1, bufsize=bufsize,
                                  stdin=green_subprocess.PIPE,
                                  stdout=green_subprocess.PIPE,
                                  stderr=green_subprocess.PIPE)
    return proc.stdin, proc.stdout, proc.stderr


@functools.wraps(original_os_popen4)
def green_popen4(cmd, mode='r', bufsize=-1):
    proc = green_subprocess.Popen(cmd, shell=1, bufsize=bufsize,
                                  stdin=green_subprocess.PIPE,
                                  stdout=green_subprocess.PIPE,
                                  stderr=green_subprocess.STDOUT)
    return proc.stdin, proc.stdout


@functools.wraps(original_os_system)
def green_system(cmd):
    return green_waitpid(green_subprocess.Popen(cmd, shell=1).pid, 0)[1]


def green_spawner(func):
    @functools.wraps(func)
    def green_version(mode, *args):
        if mode != os.P_WAIT:
            return func(mode, *args)

        pid = func(os.P_NOWAIT, *args)
        details = green_waitpid(pid, 0)[1]

        if details & 0xff:
            return -(details & 0xff)
        return details >> 8
    return green_version

green_spawnl = green_spawner(original_os_spawnl)
green_spawnle = green_spawner(original_os_spawnle)
green_spawnlp = green_spawner(original_os_spawnlp)
green_spawnlpe = green_spawner(original_os_spawnlpe)
green_spawnv = green_spawner(original_os_spawnv)
green_spawnve = green_spawner(original_os_spawnve)
green_spawnvp = green_spawner(original_os_spawnvp)
green_spawnvpe = green_spawner(original_os_spawnvpe)

is_pypy = sys.subversion[0].lower() == 'pypy'

patchers = {
    'fdopen': io.File.fromfd,
    'read': is_pypy and blocking_read or green_read,
    'write': is_pypy and blocking_write or green_write,
    'waitpid': green_waitpid,
    'wait': green_wait,
    'wait3': green_wait3,
    'wait4': green_wait4,
    'popen': green_popen,
    'popen2': green_popen2,
    'popen3': green_popen3,
    'popen4': green_popen4,
    'system': green_system,
    'spawnl': green_spawnl,
    'spawnle': green_spawnle,
    'spawnlp': green_spawnlp,
    'spawnlpe': green_spawnlpe,
    'spawnv': green_spawnv,
    'spawnve': green_spawnve,
    'spawnvp': green_spawnvp,
    'spawnvpe': green_spawnvpe,
}



from __future__ import absolute_import

import ssl

from ..io import sockets as gsock, ssl as gssl


def wrap_socket(sock, keyfile=None, certfile=None, server_side=False,
                cert_reqs=ssl.CERT_NONE, ssl_version=ssl.PROTOCOL_SSLv23,
                ca_certs=None, do_handshake_on_connect=True,
                suppress_ragged_eofs=True, ciphers=None):
    return gssl.SSLSocket(sock, keyfile, certfile, server_side, cert_reqs,
                          ssl_version, ca_certs, do_handshake_on_connect,
                          suppress_ragged_eofs, ciphers)


def get_server_certificate(
        addr, ssl_version=ssl.PROTOCOL_SSLv3, ca_certs=None):
    if (ca_certs is not None):
        cert_reqs = CERT_REQUIRED
    else:
        cert_reqs = CERT_NONE
    s = wrap_socket(gsock.Socket(), ssl_version=ssl_version,
                    cert_reqs=cert_reqs, ca_certs=ca_certs)
    s.connect(addr)
    dercert = s.getpeercert(True)
    s.close()
    return ssl.DER_cert_to_PEM_cert(dercert)


patchers = {
    'wrap_socket': wrap_socket,
    'get_server_certificate': get_server_certificate,
    'sslwrap_simple': wrap_socket,
    'SSLSocket': gssl.SSLSocket,
}



from __future__ import absolute_import

import errno
import functools
import _socket
import socket
import _ssl
import ssl
import sys
import time

from greenhouse import scheduler, util
from greenhouse.io import sockets as gsock


class SSLSocket(gsock.Socket):
    """socket-like object that speaks SSL/TLS

    this subclass of greenhouse.io.sockets.Socket wraps its underlying OS
    socket in an SSL context when necessary, and provides read and write
    methods over that channel

    mirrors the standard library ssl.SSLSocket api.
    """
    def __init__(self, sock, keyfile=None, certfile=None,
            server_side=False, cert_reqs=ssl.CERT_NONE,
            ssl_version=ssl.PROTOCOL_SSLv23, ca_certs=None,
            do_handshake_on_connect=True,
            suppress_ragged_eofs=True, ciphers=None):
        inner = sock
        while hasattr(getattr(inner, "_sock", None), "_sock"):
            inner = inner._sock
        self._sock = inner._sock if hasattr(inner, "_sock") else inner
        self._sock.setblocking(False)

        if ciphers is None and ssl_version != ssl._SSLv2_IF_EXISTS:
            ciphers = ssl._DEFAULT_CIPHERS

        if certfile and not keyfile:
            keyfile = certfile

        try:
            self.getpeername()
        except socket.error, exc:
            if exc.errno != errno.ENOTCONN:
                raise
            # not connected
            self._connected = False
            self._sslobj = None
        else:
            # connected, create the SSL object
            self._connected = True
            self._sslobj = _ssl.sslwrap(self._sock, server_side, keyfile,
                    certfile, cert_reqs, ssl_version, ca_certs, ciphers)

        self.keyfile = keyfile
        self.certfile = certfile
        self.cert_reqs = cert_reqs
        self.ssl_version = ssl_version
        self.ca_certs = ca_certs
        self.ciphers = ciphers
        self.do_handshake_on_connect = do_handshake_on_connect
        self.suppress_ragged_eofs = suppress_ragged_eofs

        if isinstance(inner, _socket.socket):
            self._timeout = socket.getdefaulttimeout()
        else:
            self._timeout = inner.gettimeout()
        if hasattr(inner, "_blocking"):
            self._blocking = inner._blocking
        else:
            self._blocking = True
        self._readable = util.Event()
        self._writable = util.Event()

        if do_handshake_on_connect and self._connected:
            self.do_handshake(self._timeout)

    def _clone(self):
        clone = object.__new__(SSLSocket)
        clone._sock = self._sock
        clone.keyfile = self.keyfile
        clone.certfile = self.certfile
        clone.cert_reqs = self.cert_reqs
        clone.ssl_version = self.ssl_version
        clone.ca_certs = self.ca_certs
        clone.ciphers = self.ciphers
        clone.do_handshake_on_connect = self.do_handshake_on_connect
        clone.suppress_ragged_eofs = self.suppress_ragged_eofs
        clone._timeout = self._timeout
        clone._blocking = self._blocking
        clone._connected = self._connected
        clone._sslobj = self._sslobj
        clone._readable = util.Event()
        clone._writable = util.Event()
        return clone

    def settimeout(self, timeout):
        self._timeout = timeout

    def gettimeout(self):
        return self._timeout

    def _read_attempt(self, length):
        try:
            return self._sslobj.read(length)
        except ssl.SSLError, exc:
            if (exc.args[0] == ssl.SSL_ERROR_EOF
                    and self.suppress_ragged_eofs):
                return ''
            raise

    def read(self, len=1024):
        'read up to len bytes and return them, or empty string on EOF'
        return self._with_retry(
                functools.partial(self._read_attempt, len),
                self.gettimeout())

    def write(self, data):
        'write data to the ssl channel and return the # of bytes transferred'
        return self._with_retry(
                functools.partial(self._sslobj.write, data),
                self.gettimeout())

    def getpeercert(self, binary_form=False):
        """return a representation of the peer's certificate

        returns None if no certificate is in use by the peer
        returns {} if a certificate is in use, but was not validated
        """
        return self._sslobj.peer_certificate(binary_form)

    def cipher(self):
        if not self._sslobj:
            return None
        return self._sslobj.cipher()

    def send(self, data, flags=0):
        if self._sslobj:
            if flags != 0:
                raise ValueError(
                        "non-zero flags not allowed in calls to send() on %s" %
                        self.__class__)
            try:
                count = self._sslobj.write(data)
            except ssl.SSLError, exc:
                if exc.args[0] == ssl.SSL_ERROR_WANT_READ:
                    return 0
                if exc.args[0] == ssl.SSL_ERROR_WANT_WRITE:
                    return 0
                raise
            return count
        else:
            return self._sock.send(data, flags)

    def sendto(self, *args):
        if self._sslobj:
            raise ValueError("sendto not allowed on instances of %s" %
                    self.__class__)
        return self._sock.sendto(*args)

    def sendall(self, data, flags=0):
        tout = _timeout(self.gettimeout())
        if self._sslobj:
            if flags != 0:
                raise ValueError(
                    "non-zero flags not allowed in calls to sendall() on %s" %
                    self.__class__)
            sent = 0
            sent = self.send(data[sent:])
            while (sent < len(data)):
                if self._blocking:
                    self._wait_event(tout.now, write=True)
                sent += self.send(data[sent:])
            return sent
        else:
            return super(SSLSocket, self).sendall(self, data, flags)

    def recv(self, buflen=1024, flags=0):
        if self._sslobj:
            if flags != 0:
                raise ValueError(
                    "non-zero flags not allowed in calls to recv() on %s" %
                    self.__class__)
            return self.read(buflen)
        else:
            return self._sock.recv(buflen, flags)

    def recv_into(self, buffer, nbytes=None, flags=0):
        if buffer and (nbytes is None):
            nbytes = len(buffer)
        elif nbytes is None:
            nbytes = 1024
        if self._sslobj:
            if flags != 0:
                raise ValueError("non-zero flags not allowed in calls to "
                        "recv_into() on %s" % self.__class__)
            temp = self.read(nbytes)
            v = len(temp)
            buffer[:v] = temp
            return v
        else:
            return self._sock.recv_into(buffer, nbytes, flags)

    def recvfrom(self, buflen=1024, flags=0):
        if self._sslobj:
            raise ValueError("recvfrom not allowed on instances of %s" %
                             self.__class__)
        else:
            return self._sock.recvfrom(buflen, flags)

    def recvfrom_into(self, buffer, nbytes=None, flags=0):
        if self._sslobj:
            raise ValueError("recvfrom_into not allowed on instances of %s" %
                             self.__class__)
        else:
            return self._sock.recvfrom_into(buffer, nbytes, flags)

    def pending(self):
        if self._sslobj:
            return self._sslobj.pending()
        else:
            return 0

    def unwrap(self):
        if self._sslobj:
            s = self._sslobj.shutdown()
            self._sslobj = None
            return s
        else:
            raise ValueError("No SSL wrapper around " + str(self))

    def shutdown(self, how):
        self._sslobj = None
        socket.shutdown(self, how)

    def close(self):
        self._sslobj = None
        self._sock = socket._closedsocket()

    def do_handshake(self, timeout):
        'perform a SSL/TLS handshake'
        tout = _timeout(timeout)
        if not self._blocking:
            return self._sslobj.do_handshake()

        while 1:
            try:
                return self._sslobj.do_handshake()
            except ssl.SSLError, exc:
                if exc.args[0] == ssl.SSL_ERROR_WANT_READ:
                    self._wait_event(tout.now)
                    continue
                elif exc.args[0] == ssl.SSL_ERROR_WANT_WRITE:
                    self._wait_event(tout.now, write=True)
                    continue
                raise

        self._wait_event(timeout)
        self._sslobj.do_handshake()

    def connect_ex(self, address):
        'connects to the address and wraps the connection in an SSL context'
        return self._connect(address, self.gettimeout())

    def _connect(self, address, timeout):
        if self._connected:
            raise ValueError("attempt to connect already-connected SSLSocket!")
        self._sslobj = _ssl.sslwrap(self._sock, False, self.keyfile,
                self.certfile, self.cert_reqs, self.ssl_version,
                self.ca_certs, self.ciphers)

        err = super(SSLSocket, self).connect_ex(address)
        if err:
            return err

        try:
            if self.do_handshake_on_connect:
                self.do_handshake(timeout)
        except socket.error, exc:
            return exc.args[0]

        self._connected = True
        return 0

    def connect(self, address):
        'connects to the address and wraps the connection in an SSL context'
        tout = _timeout(self.gettimeout())
        while 1:
            self._wait_event(tout.now, write=True)
            err = self._connect(address, tout.now)
            if err in (errno.EINPROGRESS, errno.EALREADY, errno.EWOULDBLOCK):
                continue
            if err:
                raise socket.error(err, errno.errorcode[err])
            return 0

    def accept(self):
        """accept a connection attempt from a remote client

        returns a two-tuple with the ssl-context-wrapped connection,
        and the address of the remote client
        """
        while 1:
            try:
                sock, addr = self._sock.accept()
                return (type(self)(sock,
                    keyfile=self.keyfile,
                    certfile=self.certfile,
                    server_side=True,
                    cert_reqs=self.cert_reqs,
                    ssl_version=self.ssl_version,
                    ca_certs=self.ca_certs,
                    do_handshake_on_connect=self.do_handshake_on_connect,
                    suppress_ragged_eofs=self.suppress_ragged_eofs,
                    ciphers=self.ciphers), addr)
            except socket.error, exc:
                if exc.args[0] not in (errno.EAGAIN, errno.EWOULDBLOCK):
                    raise
                sys.exc_clear()
                self._wait_event(self.gettimeout())

    def makefile(self, mode='r', bufsize=-1):
        'return a file-like object that operates on the ssl connection'
        return gsock.SocketFile(self._clone(), mode)

    def _on_readable(self):
        self._readable.set()
        self._readable.clear()

    def _on_writable(self):
        self._writable.set()
        self._writable.clear()

    def _wait_event(self, timeout=None, write=False):
        if write:
            event = self._writable
            onr, onw = None, self._on_writable
        else:
            event = self._readable
            onr, onw = self._on_readable, None

        try:
            reg = scheduler._register_fd(self.fileno(), onr, onw)
        except EnvironmentError, error:
            if error.args[0] in errno.errorcode:
                raise socket.error(*error.args)
            raise

        try:
            if event.wait(timeout):
                raise socket.timeout("timed out")
        finally:
            try:
                scheduler._unregister_fd(self.fileno(), onr, onw, reg)
            except EnvironmentError, error:
                if error.args[0] in errno.errorcode:
                    raise socket.error(*error.args)
                raise

    def _with_retry(self, func=None, timeout=None):
        if func is None:
            return lambda f: self._with_retry(f, timeout)

        tout = _timeout(timeout)
        while 1:
            try:
                return func()
            except ssl.SSLError, exc:
                if exc.args[0] == ssl.SSL_ERROR_WANT_READ:
                    self._wait_event(tout.now)
                elif exc.args[0] == ssl.SSL_ERROR_WANT_WRITE:
                    self._wait_event(tout.now, write=False)
                else:
                    raise


class _timeout(object):
    def __init__(self, timeout, exc=socket.timeout):
        if timeout is not None:
            self._deadline = time.time() + timeout
        self._timeout = timeout
        self._exc = exc

    @property
    def now(self):
        if self._timeout is None:
            return None
        timeout = self._deadline - time.time()
        if timeout < 0:
            raise self._exc('timed out')
        return timeout



import bisect
import collections
import errno
import logging
import sys
import time
import weakref

from greenhouse import compat, poller


__all__ = ["pause", "pause_until", "pause_for", "schedule", "schedule_at",
           "schedule_in", "schedule_recurring", "schedule_exception",
           "schedule_exception_at", "schedule_exception_in", "end",
           "global_exception_handler", "remove_global_exception_handler",
           "local_exception_handler", "remove_local_exception_handler",
           "handle_exception", "greenlet", "global_hook", "remove_global_hook",
           "local_incoming_hook", "remove_local_incoming_hook",
           "local_outgoing_hook", "remove_local_outgoing_hook",
           "set_ignore_interrupts", "reset_poller"]

BTREE_ORDER = 64


log = logging.getLogger("greenhouse.scheduler")

state = type('GreenhouseState', (), {})()

# from events that have triggered
state.awoken_from_events = set()

# executed a simple cooperative yield
state.paused = []

# map of file numbers to the sockets/files on that descriptor
state.descriptormap = {}

# lined up to run right away
state.to_run = collections.deque()

# exceptions queued up for scheduled coros
state.to_raise = weakref.WeakKeyDictionary()

# exception handlers, global and local
state.global_exception_handlers = []
state.local_exception_handlers = weakref.WeakKeyDictionary()

# trace hook callbacks
state.global_hooks = []
state.local_to_hooks = weakref.WeakKeyDictionary()
state.local_from_hooks = weakref.WeakKeyDictionary()

# tracks interrupts
state.interrupted = False
state.ignore_interrupts = False


class TimeoutManager(object):
    def __nonzero__(self):
        return bool(self.data)

    def first(self):
        if self.data:
            return iter(self.data).next()
        return None

    @classmethod
    def install(cls):
        state.timed_paused = cls(state.timed_paused.dump())


class BisectingTimeoutManager(TimeoutManager):
    def __init__(self, data=None):
        self.data = data or []

    def clear(self):
        del self.data[:]

    def insert(self, unixtime, glet):
        bisect.insort(self.data, (unixtime, glet))

    def check(self):
        index = bisect.bisect(self.data, (time.time(), None))
        state.to_run.extend(pair[1] for pair in self.data[:index])
        self.data = self.data[index:]

    def remove(self, unixtime, glet):
        index = bisect.bisect(self.data, (unixtime, None))
        while index < len(self.data) and self.data[index][0] == unixtime:
            if self.data[index][1] is glet:
                del self.data[index:index + 1]
                return True
            index += 1
        return False

    def dump(self):
        return self.data


class BTreeTimeoutManager(TimeoutManager):
    def __init__(self, data=None, order=BTREE_ORDER):
        self.data = btree.sorted_btree.bulkload(data or [], order)

    def clear(self):
        self.data = btree.sorted_btree(self.data.order)

    def insert(self, unixtime, glet):
        self.data.insert((unixtime, glet))

    def check(self):
        left, right = self.data.split((time.time(), None))
        state.to_run.extend(pair[1] for pair in left)
        self.data = right

    def remove(self, unixtime, glet):
        try:
            self.data.remove((unixtime, glet))
        except ValueError:
            return False
        return True

    def dump(self):
        return list(self.data)

# cooperatively yielded for a set timeout
try:
    import btree
    state.timed_paused = BTreeTimeoutManager()
except ImportError:
    state.timed_paused = BisectingTimeoutManager()


def _hit_poller(timeout):
    try:
        events = state.poller.poll(timeout)
    except KeyboardInterrupt, exc:
        # on Ctrl-C, wake up the main without killing the mainloop
        state.to_raise[compat.main_greenlet] = exc
        state.to_run.append(compat.main_greenlet)
        return
    except EnvironmentError, exc:
        if exc.args[0] != errno.EINTR:
            raise

        # interrupted by a signal
        if state.ignore_interrupts:
            events = []
        else:
            state.interrupted = True
            events = [(fd, state.poller.ERRMASK)
                      for fd in state.poller._registry.iterkeys()]

    for fd, eventmap in events:
        readables, writables = state.descriptormap.get(fd, ([], []))

        if eventmap & (state.poller.INMASK | state.poller.ERRMASK):
            for readable in readables:
                readable()

        if eventmap & (state.poller.OUTMASK | state.poller.ERRMASK):
            for writable in writables:
                writable()

    state.to_run.extend(state.awoken_from_events)
    state.awoken_from_events.clear()

    state.timed_paused.check()

    state.to_run.extend(state.paused)
    state.paused = []


def _register_fd(fd, readable, writable):
    poller = state.poller
    mask = poller.ERRMASK
    if readable:
        mask |= poller.INMASK
    if writable:
        mask |= poller.OUTMASK
    reg = state.poller.register(fd, mask)

    if fd not in state.descriptormap:
        state.descriptormap[fd] = (set(), set())
    readables, writables = state.descriptormap[fd]

    if readable:
        readables.add(readable)
    if writable:
        writables.add(writable)

    return reg


def _unregister_fd(fd, readable, writable, reg):
    if fd not in state.descriptormap:
        return
    readables, writables = state.descriptormap[fd]

    readables.discard(readable)
    writables.discard(writable)

    if not (readables or writables):
        state.descriptormap.pop(fd)

    state.poller.unregister(fd, reg)


def greenlet(func, args=(), kwargs=None):
    """create a new greenlet from a function and arguments

    :param func: the function the new greenlet should run
    :type func: function
    :param args: any positional arguments for the function
    :type args: tuple
    :param kwargs: any keyword arguments for the function
    :type kwargs: dict or None

    the only major difference between this function and that of the basic
    greenlet api is that this one sets the new greenlet's parent to be the
    greenhouse main loop greenlet, which is a requirement for greenlets that
    will wind up in the greenhouse scheduler.
    """
    if args or kwargs:
        def target():
            return func(*args, **(kwargs or {}))
    else:
        target = func
    return compat.greenlet(target, state.mainloop)


def pause():
    """pause and reschedule the current greenlet and switch to the next

    .. note:: this method blocks for a short period of time
    """
    schedule(compat.getcurrent())
    state.mainloop.switch()


def pause_until(unixtime):
    """pause and reschedule the current greenlet until a set time

    .. note:: this method will block the current greenlet

    :param unixtime: the unix timestamp of when to bring this greenlet back
    :type unixtime: int or float
    """
    schedule_at(unixtime, compat.getcurrent())
    state.mainloop.switch()


def pause_for(secs):
    """pause and reschedule the current greenlet for a set number of seconds

    .. note:: this method will block the current greenlet

    :param secs: number of seconds to pause
    :type secs: int or float
    """
    pause_until(time.time() + secs)


def schedule(target=None, args=(), kwargs=None):
    """insert a greenlet into the scheduler

    If provided a function, it is wrapped in a new greenlet

    :param target: what to schedule
    :type target: function or greenlet
    :param args:
        arguments for the function (only used if ``target`` is a function)
    :type args: tuple
    :param kwargs:
        keyword arguments for the function (only used if ``target`` is a
        function)
    :type kwargs: dict or None

    :returns: the ``target`` argument

    This function can also be used as a decorator, either preloading ``args``
    and/or ``kwargs`` or not:

    >>> @schedule
    >>> def f():
    ...     print 'hello from f'

    >>> @schedule(args=('world',))
    >>> def f(name):
    ...     print 'hello %s' % name
    """
    if target is None:
        def decorator(target):
            return schedule(target, args=args, kwargs=kwargs)
        return decorator
    if isinstance(target, compat.greenlet) or target is compat.main_greenlet:
        glet = target
    else:
        glet = greenlet(target, args, kwargs)
    state.paused.append(glet)
    return target


def schedule_at(unixtime, target=None, args=(), kwargs=None):
    """insert a greenlet into the scheduler to be run at a set time

    If provided a function, it is wrapped in a new greenlet

    :param unixtime:
        the unix timestamp at which the new greenlet should be started
    :type unixtime: int or float
    :param target: what to schedule
    :type target: function or greenlet
    :param args:
        arguments for the function (only used if ``target`` is a function)
    :type args: tuple
    :param kwargs:
        keyword arguments for the function (only used if ``target`` is a
        function)
    :type kwargs: dict or None

    :returns: the ``target`` argument

    This function can also be used as a decorator:

    >>> @schedule_at(1296423834)
    >>> def f():
    ...     print 'hello from f'

    and args/kwargs can also be preloaded:

    >>> @schedule_at(1296423834, args=('world',))
    >>> def f(name):
    ...     print 'hello %s' % name
    """
    if target is None:
        def decorator(target):
            return schedule_at(unixtime, target, args=args, kwargs=kwargs)
        return decorator
    if isinstance(target, compat.greenlet) or target is compat.main_greenlet:
        glet = target
    else:
        glet = greenlet(target, args, kwargs)
    state.timed_paused.insert(unixtime, glet)
    return target


def schedule_in(secs, target=None, args=(), kwargs=None):
    """insert a greenlet into the scheduler to run after a set time

    If provided a function, it is wrapped in a new greenlet

    :param secs: the number of seconds to wait before running the target
    :type unixtime: int or float
    :param target: what to schedule
    :type target: function or greenlet
    :param args:
        arguments for the function (only used if ``target`` is a function)
    :type args: tuple
    :param kwargs:
        keyword arguments for the function (only used if ``target`` is a
        function)
    :type kwargs: dict or None

    :returns: the ``target`` argument

    This function can also be used as a decorator:

    >>> @schedule_in(30)
    >>> def f():
    ...     print 'hello from f'

    and args/kwargs can also be preloaded:

    >>> @schedule_in(30, args=('world',))
    >>> def f(name):
    ...     print 'hello %s' % name
    """
    return schedule_at(time.time() + secs, target, args, kwargs)


def schedule_recurring(interval, target=None, maxtimes=0, starting_at=0,
                       args=(), kwargs=None):
    """insert a greenlet into the scheduler to run regularly at an interval

    If provided a function, it is wrapped in a new greenlet

    :param interval: the number of seconds between invocations
    :type interval: int or float
    :param target: what to schedule
    :type target: function or greenlet
    :param maxtimes: if provided, do not run more than ``maxtimes`` iterations
    :type maxtimes: int
    :param starting_at:
        the unix timestamp of when to schedule it for the first time (defaults
        to the time of the ``schedule_recurring`` call)
    :type starting_at: int or float
    :param args: arguments for the function
    :type args: tuple
    :param kwargs: keyword arguments for the function
    :type kwargs: dict or None

    :returns: the ``target`` argument

    This function can also be used as a decorator:

    >>> @schedule_recurring(30)
    >>> def f():
    ...     print "the regular 'hello' from f"

    and args/kwargs can also be preloaded:

    >>> @schedule_recurring(30, args=('world',))
    >>> def f(name):
    ...     print 'the regular hello %s' % name
    """
    starting_at = starting_at or time.time()

    if target is None:
        def decorator(target):
            return schedule_recurring(
                interval, target, maxtimes, starting_at, args, kwargs)
        return decorator

    func = target
    if isinstance(target, compat.greenlet) or target is compat.main_greenlet:
        if target.dead:
            raise TypeError("can't schedule a dead greenlet")
        func = target.run

    def run_and_schedule_one(tstamp, count):
        # pass in the time scheduled instead of just checking
        # time.time() so that delays don't add up
        if not maxtimes or count < maxtimes:
            tstamp += interval
            func(*args, **(kwargs or {}))
            schedule_at(tstamp, run_and_schedule_one,
                        args=(tstamp, count + 1))

    firstrun = starting_at + interval
    schedule_at(firstrun, run_and_schedule_one, args=(firstrun, 0))

    return target


def schedule_exception(exception, target):
    """schedule a greenlet to have an exception raised in it immediately

    :param exception: the exception to raise in the greenlet
    :type exception: Exception
    :param target: the greenlet that should receive the exception
    :type target: greenlet
    """
    if not isinstance(target, compat.greenlet):
        raise TypeError("can only schedule exceptions for greenlets")
    if target.dead:
        raise ValueError("can't send exceptions to a dead greenlet")
    schedule(target)
    state.to_raise[target] = exception


def schedule_exception_at(unixtime, exception, target):
    """schedule a greenlet to have an exception raised at a unix timestamp

    :param unixtime: when to raise the exception in the target
    :type unixtime: int or float
    :param exception: the exception to raise in the greenlet
    :type exception: Exception
    :param target: the greenlet that should receive the exception
    :type target: greenlet
    """
    if not isinstance(target, compat.greenlet):
        raise TypeError("can only schedule exceptions for greenlets")
    if target.dead:
        raise ValueError("can't send exceptions to a dead greenlet")
    schedule_at(unixtime, target)
    state.to_raise[target] = exception


def schedule_exception_in(secs, exception, target):
    """schedule a greenlet receive an exception after a number of seconds

    :param secs: the number of seconds to wait before raising
    :type secs: int or float
    :param exception: the exception to raise in the greenlet
    :type exception: Exception
    :param target: the greenlet that should receive the exception
    :type target: greenlet
    """
    schedule_exception_at(time.time() + secs, exception, target)


def end(target):
    """schedule a greenlet to be stopped immediately

    :param target: the greenlet to end
    :type target: greenlet
    """
    if not isinstance(target, compat.greenlet):
        raise TypeError("argument must be a greenlet")
    if not target.dead:
        schedule(target)
        state.to_raise[target] = compat.GreenletExit()


def _remove_timer(waketime, glet):
    if state.timed_paused.remove(waketime, glet):
        return True

    try:
        state.to_run.remove(glet)
    except ValueError:
        return False
    return True


@compat.greenlet
def mainloop():
    target = None
    while 1:
        # python shutdown
        if not (sys and state):
            break

        state.interrupted = False

        if not state.to_run:
            _hit_poller(0)
            while not state.to_run:
                # if there are timed-paused greenlets, we can
                # just wait until the first of them wakes up
                if state.timed_paused:
                    until = state.timed_paused.first()[0] + 0.001
                    _hit_poller(until - time.time())
                else:
                    _hit_poller(None)

        prev = target
        target = state.to_run.popleft()

        # global trace hooks
        if state.global_hooks:
            _run_global_hooks(prev, target)

        # local trace incoming hooks
        if target in state.local_to_hooks:
            _run_local_hooks(
                target, state.local_to_hooks[target], True)

        try:
            # pick up any exception we are supposed to throw in
            if target in state.to_raise:
                target.throw(state.to_raise.pop(target))
            else:
                target.switch()
        except Exception:
            # python shutdown
            if not (sys and state):
                break
            klass, exc, tb = sys.exc_info()
            handle_exception(klass, exc, tb, coro=target)
            del klass, exc, tb

        # local trace outgoing hooks
        if target in state.local_from_hooks:
            _run_local_hooks(
                target, state.local_from_hooks[target], False)

state.mainloop = mainloop


def _run_local_hooks(target, hooks, incoming):
    replacement_hooks = []
    direction = 1 if incoming else 2
    for weak in hooks:
        func = weak()
        if func is None:
            continue

        try:
            func(direction, target)
        except Exception:
            continue

        replacement_hooks.append(weak)

    hooks[:] = replacement_hooks


def _run_global_hooks(coming_from, going_to):
    replacement_hooks = []
    for weak in state.global_hooks:
        func = weak()
        if func is None:
            continue

        try:
            func(coming_from, going_to)
        except Exception:
            continue

        replacement_hooks.append(weak)

    state.global_hooks[:] = replacement_hooks


def handle_exception(klass, exc, tb, coro=None):
    """run all the registered exception handlers

    the first 3 arguments to this function match the output of
    ``sys.exc_info()``

    :param klass: the exception klass
    :type klass: type
    :param exc: the exception instance
    :type exc: Exception
    :param tb: the traceback object
    :type tb: Traceback
    :param coro:
        behave as though the exception occurred in this coroutine (defaults to
        the current coroutine)
    :type coro: greenlet

    exception handlers run would be all those added with
    :func:`global_exception_handler`, and any added for the relevant coroutine
    with :func:`local_exception_handler`.
    """
    if coro is None:
        coro = compat.getcurrent()

    replacement = []
    for weak in state.local_exception_handlers.get(coro, ()):
        func = weak()
        if func is None:
            continue

        try:
            func(klass, exc, tb)
        except Exception:
            continue

        replacement.append(weak)

    if replacement:
        state.local_exception_handlers[coro][:] = replacement

    replacement = []
    for weak in state.global_exception_handlers:
        func = weak()
        if func is None:
            continue

        try:
            func(klass, exc, tb)
        except Exception:
            continue

        replacement.append(weak)

    state.global_exception_handlers[:] = replacement


def global_exception_handler(handler):
    """add a callback for when an exception goes uncaught in any greenlet

    :param handler:
        the callback function. must be a function taking 3 arguments:

        - ``klass`` the exception class
        - ``exc`` the exception instance
        - ``tb`` the traceback object
    :type handler: function

    Note also that the callback is only held by a weakref, so if all other refs
    to the function are lost it will stop handling greenlets' exceptions
    """
    if not hasattr(handler, "__call__"):
        raise TypeError("exception handlers must be callable")
    log.info("setting a new global exception handler")

    state.global_exception_handlers.append(weakref.ref(handler))

    return handler


def remove_global_exception_handler(handler):
    """remove a callback from the list of global exception handlers

    :param handler:
        the callback, previously added via :func:`global_exception_handler`,
        to remove
    :type handler: function

    :returns: bool, whether the handler was found (and therefore removed)
    """
    for i, cb in enumerate(state.global_exception_handlers):
        cb = cb()
        if cb is not None and cb is handler:
            state.global_exception_handlers.pop(i)
            log.info("removing a global exception handler")
            return True
    return False


def local_exception_handler(handler=None, coro=None):
    """add a callback for when an exception occurs in a particular greenlet

    :param handler:
        the callback function, must be a function taking 3 arguments:

        - ``klass`` the exception class
        - ``exc`` the exception instance
        - ``tb`` the traceback object
    :type handler: function
    :param coro:
        the coroutine for which to apply the exception handler (defaults to the
        current coroutine)
    :type coro: greenlet
    """
    if handler is None:
        return lambda h: local_exception_handler(h, coro)

    if not hasattr(handler, "__call__"):
        raise TypeError("exception handlers must be callable")

    if coro is None:
        coro = compat.getcurrent()

    log.info("setting a new coroutine local exception handler")

    state.local_exception_handlers.setdefault(coro, []).append(
        weakref.ref(handler))

    return handler


def remove_local_exception_handler(handler, coro=None):
    """remove a callback from the list of exception handlers for a coroutine

    :param handler: the callback to remove
    :type handler: function
    :param coro: the coroutine for which to remove the local handler
    :type coro: greenlet

    :returns: bool, whether the handler was found (and therefore removed)
    """
    if coro is None:
        coro = compat.getcurrent()

    for i, cb in enumerate(state.local_exception_handlers.get(coro, [])):
        cb = cb()
        if cb is not None and cb is handler:
            state.local_exception_handlers[coro].pop(i)
            log.info("removing a coroutine local exception handler")
            return True
    return False


def global_hook(handler):
    """add a callback to run in every switch between coroutines

    :param handler:
        the callback function, must be a function taking 2 arguments:

        - the greenlet being switched from
        - the greenlet being switched to

        be aware that only a weak reference to this function will be held.
    :type handler: function
    """
    if not hasattr(handler, "__call__"):
        raise TypeError("trace hooks must be callable")

    log.info("setting a new global hook callback")
    state.global_hooks.append(weakref.ref(handler))

    return handler


def remove_global_hook(handler):
    """remove a callback from the list of global hooks

    :param handler:
        the callback function, previously added with global_hook, to remove
        from the list of global hooks
    :type handler: function

    :returns: bool, whether the handler was removed from the global hooks
    """
    for i, cb in enumerate(state.global_hooks):
        cb = cb()
        if cb is not None and cb is handler:
            state.global_hooks.pop(i)
            log.info("removing a global hook callback")
            return True
    return False


def local_incoming_hook(handler=None, coro=None):
    """add a callback to run every time a greenlet is about to be switched to

    :param handler:
        the callback function, must be a function taking 2 arguments:

        - an integer indicating whether it is being called as an incoming (1)
          hook or an outgoing (2) hook (in this case it will always receive 1).
        - the coroutine being switched into (in this case it will always be the
          same as the one indicated by the ``coro`` argument to
          ``local_incoming_hook``.

        Be aware that only a weak reference to this function will be held.
    :type handler: function
    :param coro:
        the coroutine for which to apply the trace hook (defaults to current)
    :type coro: greenlet
    """
    if handler is None:
        return lambda h: local_incoming_hook(h, coro)

    if not hasattr(handler, "__call__"):
        raise TypeError("trace hooks must be callable")

    if coro is None:
        coro = compat.getcurrent()

    log.info("setting a coroutine incoming local hook callback")

    state.local_to_hooks.setdefault(coro, []).append(
        weakref.ref(handler))

    return handler


def remove_local_incoming_hook(handler, coro=None):
    """remove a callback from the incoming hooks for a particular coro

    :param handler: the callback previously added via local_incoming_hook
    :type handler: function
    :param coro: the coroutine for which the hook should be removed
    :type coro: greenlet

    :returns: bool, whether the handler was found and removed
    """
    if coro is None:
        coro = compat.getcurrent()

    for i, cb in enumerate(state.local_to_hooks.get(coro, [])):
        cb = cb()
        if cb is not None and cb is handler:
            log.info("removing a coroutine incoming local hook callback")
            state.local_to_hooks[coro].pop(i)
            return True
    return False


def local_outgoing_hook(handler=None, coro=None):
    """add a callback to run every time a greenlet is switched away from

    :param handler:
        the callback function, must be a function taking 2 arguments:

        - an integer indicating whether it is being called as an incoming (1)
          hook or as an outgoing (2) hook (in this case it will always be 2).
        - the coroutine being switched from (in this case it is the one
          indicated by the ``coro`` argument to ``local_outgoing_hook``.

        Be aware that only a weak reference to this function will be held.
    :type handler: function
    :param coro:
        the coroutine for which to apply the trace hook (defaults to current)
    :type coro: greenlet
    """
    if handler is None:
        return lambda h: local_outgoing_hook(h, coro)

    if not hasattr(handler, "__call__"):
        raise TypeError("trace hooks must be callable")

    if coro is None:
        coro = compat.getcurrent()

    log.info("setting a coroutine local outgoing hook callback")

    state.local_from_hooks.setdefault(coro, []).append(
        weakref.ref(handler))

    return handler


def remove_local_outgoing_hook(handler, coro=None):
    """remove a callback from the outgoing hooks for a particular coro

    :param handler: the callback previously added via local_outgoing_hook
    :type handler: function
    :param coro: the coroutine for which the hook should be removed
    :type coro: greenlet

    :returns: bool, whether the handler was found and removed
    """
    if coro is None:
        coro = compat.getcurrent()

    for i, cb in enumerate(state.local_from_hooks.get(coro, [])):
        cb = cb()
        if cb is not None and cb is handler:
            log.info("removing a coroutine outgoing local hook callback")
            state.local_from_hooks[coro].pop(i)
            return True
    return False


def set_ignore_interrupts(flag=True):
    """turn off EINTR-raising from emulated syscalls on interruption by signals

    due to the nature of greenhouse's system call emulation,
    ``signal.siginterrupt`` can't be made to work with it. specifically,
    greenhouse can't differentiate between different signals. so this function
    toggles whether to restart for *ALL* or *NO* signals.

    :param flag:
        whether to turn EINTR exceptions off (``True``) or on (``False``)
    :type flag: bool
    """
    log.info("setting ignore_interrupts to %r" % flag)
    state.ignore_interrupts = bool(flag)


def reset_poller(poll=None):
    """replace the scheduler's poller, throwing away any pre-existing state

    this is only really a good idea in the new child process after a fork(2).
    """
    state.poller = poll or poller.best()
    log.info("resetting fd poller, using %s" % type(state.poller).__name__)



from __future__ import absolute_import

import errno
import fcntl
import functools

from .. import scheduler


TIMEOUT = 0.01

original_flock = fcntl.flock
original_lockf = fcntl.lockf
original_fcntl = fcntl.fcntl


@functools.wraps(original_flock)
def green_flock(fd, operation):
    # pass unlocks and non-blocking acquires straight through
    if operation & (fcntl.LOCK_NB | fcntl.LOCK_UN):
        return original_flock(fd, operation)

    # doing a blocking lock acquire - fake the blocking
    while 1:
        try:
            return original_flock(fd, operation | fcntl.LOCK_NB)
        except EnvironmentError, exc:
            if exc.args[0] not in (errno.EACCES, errno.EAGAIN):
                raise
            scheduler.pause_for(TIMEOUT)


@functools.wraps(original_lockf)
def green_lockf(fd, operation, length=0, start=0, whence=0):
    # pass unlocks and non-blocking acquires straight through
    if operation & (fcntl.LOCK_NB | fcntl.LOCK_UN):
        return original_lockf(fd, operation, length, start, whence)

    # doing a blocking lock acquire - fake the blocking
    while 1:
        try:
            return original_lockf(fd, operation | fcntl.LOCK_NB,
                                  length, start, whence)
        except EnvironmentError, exc:
            if exc.args[0] not in (errno.EACCES, errno.EAGAIN):
                raise
            scheduler.pause_for(TIMEOUT)


@functools.wraps(original_fcntl)
def green_fcntl(fd, opt, arg=0):
    # pass non-blocking calls straight through
    if opt != fcntl.F_SETLKW:
        return original_fcntl(fd, opt, arg)

    # doing a blocking lock acquire - fake the blocking
    while 1:
        try:
            return original_fcntl(fd, fcntl.F_SETLK, arg)
        except EnvironmentError, exc:
            if exc.args not in (errno.EACCES, errno.EAGAIN):
                raise
            scheduler.pause_for(TIMEOUT)


patchers = {
    'flock': green_flock,
    'lockf': green_lockf,
    'fcntl': green_fcntl,
}



"""monkey-patching facilities for greenhouse's  cooperative stdlib replacements

many of the apis in greenhouse (particularly in the :mod:`greenhouse.util`
module) have been explicitly made to match the signatures and behavior of I/O
and threading related apis from the python standard library.

this module enables monkey-patching the stdlib modules to swap in the
greenhouse versions, the idea being that you can cause a third-party library to
use coroutines without it having to be explicitly written that way.
"""
from __future__ import absolute_import, with_statement

import contextlib
import sys
import logging
import types

from .. import io, scheduler, util


__all__ = ["patch", "unpatch", "patched", "patched_context"]

log = logging.getLogger("greenhouse.emulation")


def patched(module_name):
    """import and return a named module with patches applied locally only

    this function returns a module after importing it in such as way that it
    will operate cooperatively, but not overriding the module globally.

    >>> green_httplib = patched("httplib")
    >>> # using green_httplib will only block greenlets
    >>> import httplib
    >>> # using httplib will block threads/processes
    >>> # both can exist simultaneously

    :param module_name:
        the module's name that is to be imported. this can be a dot-delimited
        name, in which case the module at the end of the path is the one that
        will be returned
    :type module_name: str

    :returns:
        the module indicated by module_name, imported so that it will not block
        globally, but also not touching existing global modules
    """
    if module_name in _patchers:
        return _patched_copy(module_name, _patchers[module_name])

    # grab the unpatched version of the module for posterity
    old_module = sys.modules.pop(module_name, None)

    # apply all the standard library patches we have
    saved = [(module_name, old_module)]
    for name, patch in _patchers.iteritems():
        new_mod = _patched_copy(name, patch)
        saved.append((name, sys.modules.pop(name)))
        sys.modules[name] = new_mod

    try:
        # import the requested module with patches in place
        result = __import__(module_name, {}, {}, module_name.rsplit(".", 1)[0])
    finally:
        # put all the original modules back as they were
        for name, old_mod in saved:
            if old_mod is None:
                sys.modules.pop(name, None)
            else:
                sys.modules[name] = old_mod

    return result


@contextlib.contextmanager
def patched_context(*module_names, **kwargs):
    """apply emulation patches only for a specific context

    :param module_names: var-args for the modules to patch, as in :func:`patch`
    :param local:
        if True, unpatching is done on every switch-out, and re-patching on
        every switch-in, so that they are only applied for the one coroutine

    :returns:
        a contextmanager that patches on ``__enter__`` and unpatches on
        ``__exit__``
    """
    local = kwargs.pop('local', False)
    if kwargs:
        raise TypeError("patched_context() got an unexpected keyword " +
                        "argument %r" % kwargs.keys()[0])

    patch(*module_names)
    if local:
        @scheduler.local_incoming_hook
        @scheduler.local_outgoing_hook
        def hook(direction, target):
            {1: patch, 2: unpatch}[direction](*module_names)

    yield

    unpatch(*module_names)
    if local:
        scheduler.remove_local_incoming_hook(hook)
        scheduler.remove_local_outgoing_hook(hook)


def _patched_copy(mod_name, patch):
    old_mod = __import__(mod_name, {}, {}, mod_name.rsplit(".", 1)[0])
    new_mod = types.ModuleType(old_mod.__name__)
    new_mod.__dict__.update(old_mod.__dict__)
    new_mod.__dict__.update(patch)
    return new_mod


# the definitive list of which attributes of which modules get monkeypatched
# (this gets added to with submodule imports below)
_patchers = {
    '__builtin__': {
        'file': io.File,
        'open': io.File,
    },

    'Queue': {
        'Queue': util.Queue,
        'LifoQueue': util.LifoQueue,
        'PriorityQueue': util.PriorityQueue,
    },

    'sys': {
        'stdin': io.files.stdin,
        'stdout': io.files.stdout,
        'stderr': io.files.stderr,
    },

    'time': {
        'sleep': scheduler.pause_for,
    }
}


def patch(*module_names):
    """apply monkey-patches to stdlib modules in-place

    imports the relevant modules and simply overwrites attributes on the module
    objects themselves. those attributes may be functions, classes or other
    attributes.

    valid arguments are:

    - __builtin__
    - Queue
    - fcntl
    - os
    - select
    - signal
    - socket
    - ssl
    - sys
    - thread
    - threading
    - time
    - zmq

    with no arguments, patches everything it can in all of the above modules

    :raises: ``ValueError`` if an unknown module name is provided

    .. note::
        lots more standard library modules can be made non-blocking by virtue
        of patching some combination of the the above (because they only block
        by using blocking functions defined elsewhere). a few examples:

        - ``subprocess`` works cooperatively with ``os`` and ``select`` patched
        - ``httplib``, ``urllib`` and ``urllib2`` will all operate
          cooperatively with ``socket`` and ``ssl`` patched
    """
    if not module_names:
        module_names = _patchers.keys()

    log.info("monkey-patching in-place (%d modules)" % len(module_names))

    for module_name in module_names:
        if module_name not in _patchers:
            raise ValueError("'%s' is not greenhouse-patchable" % module_name)

    for module_name in module_names:
        if module_name in sys.modules:
            module = sys.modules[module_name]
        else:
            module = __import__(
                module_name, {}, {}, module_name.rsplit(".", 1)[0])
        for attr, patch in _patchers[module_name].items():
            setattr(module, attr, patch)


def unpatch(*module_names):
    """undo :func:`patch`\es to standard library modules

    this function takes one or more module names and puts back their patched
    attributes to the standard library originals.

    valid arguments are the same as for :func:`patch`.

    with no arguments, undoes all monkeypatches that have been applied

    :raises: ``ValueError`` if an unknown module name is provided
    """
    if not module_names:
        module_names = _standard.keys()

    log.info("undoing monkey-patches in-place (%d modules)" %
             len(module_names))

    for module_name in module_names:
        if module_name not in _standard:
            raise ValueError("'%s' is not greenhouse-patchable" % module_name)

    for module_name in module_names:
        module = __import__(module_name, {}, {}, module_name.rsplit(".", 1)[0])
        for attr, value in _standard[module_name].items():
            setattr(module, attr, value)


from . import signal
_patchers['signal'] = signal.patchers

from . import select
_patchers['select'] = select.patchers

from . import socket
_patchers['socket'] = socket.patchers

from . import ssl
_patchers['ssl'] = ssl.patchers

from . import threading
_patchers['thread'] = threading.thread_patchers
_patchers['threading'] = threading.threading_patchers

from . import os
_patchers['os'] = os.patchers
# implementing child process-related
# things in the os module in terms of this
os.green_subprocess = patched('subprocess')

from . import fcntl
_patchers['fcntl'] = fcntl.patchers

from . import zmq
if zmq.patchers:
    _patchers['zmq'] = zmq.patchers


if zmq.backend_patchers:
    _patchers['zmq.backend'] = zmq.backend_patchers
if zmq.sugar_patchers:
    _patchers['zmq.sugar'] = zmq.sugar_patchers
if zmq.sugar_context_patchers:
    _patchers['zmq.sugar.context'] = zmq.sugar_context_patchers
if zmq.sugar_socket_patchers:
    _patchers['zmq.sugar.socket'] = zmq.sugar_socket_patchers
if zmq.sugar_poll_patchers:
    _patchers['zmq.sugar.poll'] = zmq.sugar_poll_patchers

try:
    from . import dns
    _patchers['socket'].update(dns.socket_patchers)
except ImportError:
    pass


_standard = {}
for mod_name, patchers in _patchers.items():
    _standard[mod_name] = {}

    module = __import__(mod_name, {}, {}, mod_name.rsplit(".", 1)[0])
    for attr_name, patcher in patchers.items():
        _standard[mod_name][attr_name] = getattr(module, attr_name, None)
del mod_name, patchers, module, attr_name, patcher



from __future__ import absolute_import

# yes, this is an unused import, but stdlib threading.py checks
# `'dummy_import' in sys.modules` when it's deleting the main thread, and
# doesn't re-raise a KeyError in that case. so we just need to ensure that
# dummy_threading gets imported.
import dummy_threading

import thread

from .. import compat, scheduler, util


def green_start(function, args, kwargs=None):
    glet = scheduler.greenlet(function, args, kwargs)
    scheduler.schedule(glet)
    return id(glet)


def thread_exit():
    raise SystemExit()


def thread_get_ident():
    return current_thread().ident


def thread_stack_size(size=None):
    if size is not None:
        raise thread.ThreadError()
    # doesn't really apply, but whatever
    return thread.stack_size()

_main_thread = object.__new__(util.Thread)
_main_thread.__dict__.update({
    'name': 'MainThread',
    '_target': None,
    '_args': (),
    '_kwargs': {},
    '_started': True,
    '_finished': util.Event(),
    '_glet': compat.main_greenlet,
    '_ident': id(compat.main_greenlet),
})
_main_thread._activate()


class _DummyThread(util.Thread):
    def __init__(self):
        super(_DummyThread, self).__init__(name=self._newname("Dummy-%d"))

    def join(self, timeout=None):
        # this is a strange way for the stdlib to raise here,
        # but we'll just match the raised exception
        assert False, "cannot join a dummy thread"


def enumerate_threads():
    return util.Thread._active.values()


def active_thread_count():
    return len(util.Thread._active)


def current_thread():
    glet = compat.getcurrent()
    if glet in util.Thread._active:
        return util.Thread._active[glet]

    thread = _DummyThread()
    thread._glet = glet
    thread._ident = id(glet)
    thread._started = True
    return thread


thread_patchers = {
    'allocate_lock': util.Lock,
    'allocate': util.Lock,
    'exit': thread_exit,
    'exit_thread': thread_exit,
    'get_ident': thread_get_ident,
    '_local': util.Local,
    'LockType': util.Lock,
    'stack_size': thread_stack_size,
    'start_new_thread': green_start,
    'start_new': green_start,
}


threading_patchers = {
    'Event': util.Event,
    'Lock': util.Lock,
    'RLock': util.RLock,
    'Condition': util.Condition,
    'Semaphore': util.Semaphore,
    'BoundedSemaphore': util.BoundedSemaphore,
    'Timer': util.Timer,
    'Thread': util.Thread,
    'local': util.Local,
    'enumerate': enumerate_threads,
    'active_count': active_thread_count,
    'activeCount': active_thread_count,
    'current_thread': current_thread,
    'currentThread': current_thread,
    '_allocate_lock': util.Lock,
    '_sleep': scheduler.pause_for,
    '_start_new_thread': green_start,
    '_active': util.Thread._active_by_id,
}

threading_local_patchers = {
    'RLock': util.RLock,
    'current_thread': current_thread,
    'local': util.Local,
}



"""
These functions enable running an additional server in greenhouse server
processes, which accepts connections and runs interactive python interpreters
on them, enabling entirely flexible and ad-hoc server administration at
runtime.

.. warning:: **backdoors are a gaping security hole**

    Make certain that you use ``"127.0.0.1"`` as the host on which to listen
    for connections so that it will only accept connection requests made
    locally. If you must connect to it from another machine, at least make sure
    it is behind a firewall that will block the backdoor port.
"""
from __future__ import absolute_import, with_statement

import code
import collections
import contextlib
import logging
import socket
import sys
try:
    from cStringIO import StringIO
except ImportError:
    from StringIO import StringIO

from . import io, scheduler


__all__ = ["run_backdoor", "backdoor_handler"]


log = logging.getLogger("greenhouse.backdoor")

PREAMBLE = "Python %s on %s" % (sys.version, sys.platform)
PS1 = getattr(sys, "ps1", ">>> ")
PS2 = getattr(sys, "ps2", "... ")


def run_backdoor(address, namespace=None):
    """start a server that runs python interpreters on connections made to it

    .. note::

        this function blocks effectively indefinitely -- it runs the listening
        socket loop in the current greenlet. to keep the current greenlet free,
        :func:`schedule<greenhouse.scheduler.schedule>` this function.

    :param address:
        the address on which to listen for backdoor connections, in the form of
        a two-tuple ``(host, port)``
    :type address: tuple
    :param namespace:
        the local namespace dict for the interpreter, or None to have each
        connection create its own empty namespace
    :type namespace: dict or None
    """
    log.info("starting on %r" % (address,))
    serversock = io.Socket()
    serversock.setsockopt(socket.SOL_SOCKET, socket.SO_REUSEADDR, 1)
    serversock.bind(address)
    serversock.listen(socket.SOMAXCONN)

    while 1:
        clientsock, address = serversock.accept()
        log.info("connection received from %r" % (address,))
        scheduler.schedule(backdoor_handler, args=(clientsock, namespace))


def backdoor_handler(clientsock, namespace=None):
    """start an interactive python interpreter on an existing connection

    .. note::
        this function will block for as long as the connection remains alive.

    :param sock: the socket on which to serve the interpreter
    :type sock: :class:`Socket<greenhouse.io.sockets.Socket>`
    :param namespace:
        the local namespace dict for the interpreter, or None to have the
        function create its own empty namespace
    :type namespace: dict or None
    """
    namespace = {} if namespace is None else namespace.copy()
    console = code.InteractiveConsole(namespace)
    multiline_statement = []
    stdout, stderr = StringIO(), StringIO()

    clientsock.sendall(PREAMBLE + "\n" + PS1)

    for input_line in _produce_lines(clientsock):
        input_line = input_line.rstrip()
        if input_line:
            input_line = '\n' + input_line
        source = '\n'.join(multiline_statement) + input_line
        response = ''

        with _wrap_stdio(stdout, stderr):
            result = console.runsource(source)

        response += stdout.getvalue()
        err = stderr.getvalue()
        if err:
            response += err

        if err or not result:
            multiline_statement = []
            response += PS1
        else:
            multiline_statement.append(input_line)
            response += PS2

        clientsock.sendall(response)


@contextlib.contextmanager
def _wrap_stdio(stdout, stderr):
    @scheduler.local_incoming_hook
    @scheduler.local_outgoing_hook
    def hook(direction, coro):
        if direction == 1:
            # incoming
            sys.stdout = stdout
            sys.stderr = stderr
        else:
            # outgoing
            sys.stdout = real_stdout
            sys.stderr = real_stderr

    real_stdout = sys.stdout
    real_stderr = sys.stderr

    stdout.seek(0)
    stderr.seek(0)
    stdout.truncate()
    stderr.truncate()

    hook(1, None)

    yield

    hook(2, None)
    scheduler.remove_local_incoming_hook(hook)
    scheduler.remove_local_outgoing_hook(hook)


# yields lines and closes connection on '\x04' (sent by telnet on Ctrl-D)
def _produce_lines(sock):
    lines = collections.deque()
    buf = ""

    while 1:
        if lines:
            yield lines.popleft()

        while not lines:
            block = sock.recv(8192)
            if not block:
                if buf:
                    yield buf
                log.info("EOF received, closing connection")
                return

            if block[0] == '\x04':
                sock.close()
                if buf:
                    yield buf
                log.info("Ctrl-D received, closing connection")
                return

            l = (buf + block).split("\n")
            lines.extend(l[:-1])
            buf = l[-1]



from __future__ import with_statement

import sys

from greenhouse import scheduler, util


__all__ = ["OneWayPool", "Pool", "OrderedPool", "map", "PoolClosed"]

_STOP = object()


class PoolClosed(RuntimeError):
    """Exception raised to wake any coroutines blocked on :meth:`get<Pool.get>`
    calls when a pool is :meth:`closed<Pool.close>`
    """


class OneWayPool(object):
    """a pool whose workers have input only

    :param func: the function the workers should run repeatedly
    :type func: function
    :param size: the number of workers to run
    :type size: int

    this class can be used as a context manager, in which case :meth:`start` is
    called at entry and :meth:`close` is called on exit from the context.
    """
    def __init__(self, func, size=10):
        self.func = func
        self.size = size
        self.inq = util.Queue()
        self._closing = False

    def start(self):
        "start the pool's workers"
        for i in xrange(self.size):
            scheduler.schedule(self._runner)
        self._closing = False

    def close(self):
        """shut down the pool's workers

        this method sets the :attr:`closing` attribute, and once all queued
        work has been completed it will set the :attr:`closed` attribute
        """
        self._closing = True
        for i in xrange(self.size):
            self.inq.put(_STOP)

    __del__ = close

    @property
    def closing(self):
        "whether :meth:`close` has been called"
        return self._closing

    @property
    def closed(self):
        "the pool has been closed and all :meth:`put` items have been handled"
        return self._closing and self.inq._jobs_done.is_set()

    def _runner(self):
        while 1:
            input = self.inq.get()
            if input is _STOP:
                self.inq.task_done()
                break
            self._handle_one(input)
            self.inq.task_done()

    def _handle_one(self, input):
        rval, success = self._run_func(*input)
        if not success:
            scheduler.handle_exception(*rval)

    def _run_func(self, args, kwargs):
        try:
            return self.func(*args, **kwargs), True
        except Exception:
            klass, exc, tb = sys.exc_info()
            return (klass, exc, tb), False

    def put(self, *args, **kwargs):
        """place a new item into the pool to be handled by the workers

        all positional and keyword arguments will be passed in as the arguments
        to the function run by the pool's workers
        """
        self.inq.put((args, kwargs))

    def join(self, timeout=None):
        """wait for the pool's input queue to be cleaned out

        .. note::
            this method will block until it has no more pending tasks or, if a
            timeout is provided, it expires

        :param timeout: maximum time in seconds to wait
        :type timeout: float or None
        """
        self.inq.join(timeout)

    def __enter__(self):
        self.start()
        return self

    def __exit__(self, klass, value, tb):
        self.close()


class Pool(OneWayPool):
    """a pool from which the function's return values can be retrieved

    this class supports use as a context manager just as :class:`OneWayPool`
    does, and it also supports iteration, in which case :meth:`get` results
    are yielded until the pool has been closed and all results pulled.
    """
    def __init__(self, *args, **kwargs):
        super(Pool, self).__init__(*args, **kwargs)
        self.outq = util.Queue()

    def __iter__(self):
        while True:
            try:
                yield self.get()
            except PoolClosed:
                return

    def _handle_one(self, input):
        self.outq.put(self._run_func(*input))

    def get(self):
        """retrieve a result from the pool

        if nothing is already completed when this method is called, it will
        block until something comes back

        if the pool's function exited via exception, that will come back as
        a result here as well, but will be re-raised in :meth:`get`.

        .. note::
            if there is nothing in the pool's output queue when this method is
            called, it will block until something is ready

        :returns:
            a return value from one of the function's invocations if it exited
            normally

        :raises:
            :class:`PoolClosed` if the pool was closed before a result could be
            produced for thie call

        :raises: any exception that was raised inside the worker function
        """
        if self.closed:
            raise PoolClosed()

        result, succeeded = self.outq.get()
        if not succeeded:
            klass, exc, tb = result
            raise klass, exc, tb
        return result

    def close(self):
        """shut down the pool's workers

        this method sets the :attr:`closing` attribute, lines up the
        :attr:`closed` attribute to be set once any queued data has been
        processed, and raises a PoolClosed() exception in any coroutines still
        blocked on :meth:`get`.
        """
        super(Pool, self).close()
        for waiter, waketime in self.outq._waiters:
            scheduler.schedule_exception(PoolClosed(), waiter)

    @property
    def closed(self):
        """the pool has been closed, all :meth:`put` items have been retrieved
        with :meth`get`."""
        return super(Pool, self).closed and self.outq.empty()


class OrderedPool(Pool):
    """a pool in which the results are produced in order

    :class:`Pool` can produce results in a different order than that in which
    they were :meth:`put<Pool.put>` if one function invocation blocks for
    longer than another, but this class enforces that nothing will be produced
    by :meth:`get` until the next one in the order in which they were
    :meth:`put`.

    this class supports context manager usage and iteration just as its parent
    class does.
    """
    def __init__(self, func, size=10):
        super(OrderedPool, self).__init__(func, size)
        self._putcount = 0
        self._getcount = 0
        self._cache = {}

    def _handle_one(self, input):
        count, input = input
        self.outq.put((count, self._run_func(*input)))

    def put(self, *args, **kwargs):
        """place a new item into the pool to be handled by the workers

        all positional and keyword arguments will be passed in as the arguments
        to the function run by the pool's workers
        """
        self.inq.put((self._putcount, (args, kwargs)))
        self._putcount += 1

    def get(self):
        """retrieve a result from the pool

        if nothing is already completed when this method is called, it will
        block until something comes back

        if the pool's function exited via exception, that will come back as
        a result here as well, but will be re-raised in :meth:`get`.

        .. note::
            if there is nothing in the pool's output queue when this method is
            called, it will block until something is ready

        :returns:
            a return value from one of the function's invocations if it exited
            normally

        :raises:
            :class:`PoolClosed` if the pool was closed before a result could be
            produced for thie call

        :raises: any exception that was raised inside the worker function
        """
        if self.closed:
            raise PoolClosed()

        while self._getcount not in self._cache:
            counter, result = self.outq.get()
            self._cache[counter] = result

        result, succeeded = self._cache.pop(self._getcount)
        self._getcount += 1

        if not succeeded:
            klass, exc, tb = result
            raise klass, exc, tb
        return result


def map(func, items, pool_size=10):
    """a parallelized work-alike to the built-in ``map`` function

    this function works by creating an :class:`OrderedPool` and placing all
    the arguments in :meth:`put<OrderedPool.put>` calls, then yielding items
    produced by the pool's :meth:`get<OrderedPool.get>` method.

    :param func: the mapper function to use
    :type func: function
    :param items: the items to use as the mapper's arguments
    :type items: iterable
    :param pool_size:
        the number of workers for the pool -- this amounts to the concurrency
        with which the map is accomplished (default 10)
    :type pool_size: int

    :returns:
        a lazy iterator (like python3's map or python2's itertools.imap) over
        the results of the mapping
    """
    with OrderedPool(func, pool_size) as pool:
        for count, item in enumerate(items):
            pool.put(item)

        for i in xrange(count + 1):
            yield pool.get()



from __future__ import absolute_import, with_statement

import contextlib
import errno
import fcntl
import os
import sys
try:
    from cStringIO import StringIO
except ImportError:
    from StringIO import StringIO

from .. import scheduler, util


__all__ = ["File", "stdin", "stdout", "stderr"]

_open = open
_file = file
_read = os.read
_write = os.write
_fdopen = os.fdopen
_osopen = os.open
_osclose = os.close
_fcntl = fcntl.fcntl


class FileBase(object):
    CHUNKSIZE = 8192
    NEWLINE = "\n"

    def __init__(self):
        self._rbuf = StringIO()
        self.encoding = None

    def __iter__(self):
        line = self.readline()
        while line:
            yield line
            line = self.readline()

    def __enter__(self):
        return self

    def __exit__(self, type, value, traceback):
        self.close()

    def read(self, size=-1):
        """read a number of bytes from the file and return it as a string

        .. note:: this method will block if there is no data already available

        :param size:
            the maximum number of bytes to read from the file. < 0 means read
            the file to the end
        :type size: int

        :returns: a string of the read file contents
        """
        chunksize = size < 0 and self.CHUNKSIZE or min(self.CHUNKSIZE, size)

        buf = self._rbuf
        buf.seek(0, os.SEEK_END)
        collected = buf.tell()

        while 1:
            if size >= 0 and collected >= size:
                # we have read enough already
                break

            output = self._read_chunk(chunksize)
            if output is None:
                continue

            if not output:
                # nothing more to read
                break

            collected += len(output)
            buf.write(output)

        # get rid of the old buffer
        rc = buf.getvalue()
        buf.seek(0)
        buf.truncate()

        if size >= 0:
            # leave the overflow in the buffer
            buf.write(rc[size:])
            return rc[:size]
        return rc

    def readline(self, max_len=-1):
        """read from the file until a newline is encountered

        .. note::

            this method will block if there isn't already a full line available
            from the data source

        :param max_len: stop reading a single line after this many bytes
        :type max_len: int

        :returns:
            a string of the line it read from the file, including the newline
            at the end
        """
        buf = self._rbuf
        newline, chunksize = self.NEWLINE, self.CHUNKSIZE
        buf.seek(0)

        text = buf.read()
        if len(text) >= max_len >= 0:
            buf.seek(0)
            buf.truncate()
            buf.write(text[max_len:])
            return text[:max_len]

        while text.find(newline) < 0:
            text = self._read_chunk(chunksize)
            if text is None:
                text = ''
                continue
            if buf.tell() + len(text) >= max_len >= 0:
                text = buf.getvalue() + text
                buf.seek(0)
                buf.truncate()
                buf.write(text[max_len:])
                return text[:max_len]
            if not text:
                break
            buf.write(text)
        else:
            # found a newline
            rc = buf.getvalue()
            index = rc.find(newline) + len(newline)

            buf.seek(0)
            buf.truncate()
            buf.write(rc[index:])
            return rc[:index]

        # hit the end of the file, no more newlines
        rc = buf.getvalue()
        buf.seek(0)
        buf.truncate()
        return rc

    def readlines(self, bufsize=-1):
        """reads the entire file, producing the lines one at a time

        .. note:: this method will block until the data source is exhausted

        :param bufsize: the read buffer size to use
        :type bufsize: int

        :returns: a list of all the lines in the file to the end
        """
        return list(self.__iter__())

    def write(self, data):
        """write data to the file

        :param data:
            the data to write into the file, at the descriptor's current
            position
        :type data: str
        """
        while data:
            went = self._write_chunk(data)
            if went is None:
                continue
            data = data[went:]

    def writelines(self, lines):
        """write a sequence of strings into the file

        :meth:`writelines` does not add newlines to the strings

        :param lines: a sequence of strings to write to the file
        :type lines: iterable
        """
        self.write("".join(lines))

    @staticmethod
    def _mode_to_flags(mode):
        flags = os.O_RDONLY | os.O_NONBLOCK  # always non-blocking
        if (('w' in mode or 'a' in mode) and 'r' in mode) or '+' in mode:
            # both read and write
            flags |= os.O_RDWR
        elif 'w' in mode or 'a' in mode:
            # else just write
            flags |= os.O_WRONLY

        if 'a' in mode:
            # append-write mode
            flags |= os.O_APPEND

        # builtin open() also creates the file on writable opens
        if flags & (os.O_WRONLY|os.O_RDWR):
            flags |= os.O_CREAT

        return flags


class File(FileBase):
    """the greenhouse drop-in replacement for the built-in ``file`` type

    this class will use non-blocking file descriptors for its IO, so that
    whatever readystate information the OS and filesystem provide will be used
    to block only a single coroutine rather than the whole process.
    unfortunately filesystems tend to be very unreliable in this regard.
    """
    def __init__(self, name, mode='rb'):
        super(File, self).__init__()
        self.mode = mode
        self.name = name
        self._closed = False

        # translate mode into the proper open flags
        flags = self._mode_to_flags(mode)

        # open the file, get a descriptor
        try:
            self._fileno = _osopen(name, flags)
        except OSError, exc:
            # stdlib open() raises IOError if the file doesn't exist, os.open
            # raises OSError. pfft, whatever.
            raise IOError(*exc.args)

        # try to drive the asyncronous waiting off of the polling interface,
        # but epoll doesn't seem to support filesystem descriptors, so fall
        # back to waiting with a simple yield
        self._set_up_waiting()

    @contextlib.contextmanager
    def _registered(self, read=True, write=True):
        rd = self._on_readable if read else None
        wr = self._on_writable if write else None
        try:
            reg = scheduler._register_fd(self._fileno, rd, wr)
        except EnvironmentError, exc:
            if exc.args and exc.args[0] in errno.errorcode:
                raise IOError, IOError(*exc.args), sys.exc_info()[2]
            raise

        try:
            yield
        finally:
            try:
                scheduler._unregister_fd(self._fileno, rd, wr, reg)
            except EnvironmentError, exc:
                if exc.args and exc.args[0] in errno.errorcode:
                    raise IOError, IOError(*exc.args), sys.exc_info()[2]
                raise

    def _set_up_waiting(self):
        if scheduler.state.poller.supports(self):
            self._wait = self._wait_event
            self._readable = util.Event()
            self._writable = util.Event()
        else:
            self._wait = self._wait_yield

    def _on_readable(self):
        self._readable.set()
        self._readable.clear()

    def _on_writable(self):
        self._writable.set()
        self._writable.clear()

    def _wait_event(self, reading):
        "wait on our events"
        with self._registered(reading, not reading):
            (self._readable if reading else self._writable).wait()

        if scheduler.state.interrupted:
            raise IOError(errno.EINTR, "interrupted system call")

    def _wait_yield(self, reading):
        "generic busy wait, for when polling won't work"
        scheduler.pause()

    def _read_chunk(self, size):
        try:
            return _read(self._fileno, size)
        except EnvironmentError, err:
            if err.args[0] in (errno.EAGAIN, errno.EINTR):
                self._wait(reading=True)
                return None
            raise

    def _write_chunk(self, data):
        try:
            return _write(self._fileno, data)
        except EnvironmentError, err:
            if err.args[0] in (errno.EAGAIN, errno.EINTR):
                self._wait(reading=False)
                return None
            raise

    @staticmethod
    def _add_flags(fd, flags):
        fdflags = _fcntl(fd, fcntl.F_GETFL)
        if fdflags | flags != fdflags:
            _fcntl(fd, fcntl.F_SETFL, flags | fdflags)

    @classmethod
    def fromfd(cls, fd, mode='rb', bufsize=-1):
        """create a cooperating greenhouse file from an existing descriptor

        :param fd: the file descriptor to wrap in a new file object
        :type fd: int
        :param mode: the file mode
        :type mode: str
        :param bufsize:
            the size of read buffer to use. 0 indicates unbuffered, and < 0
            means use the system default. defaults to -1

        :returns: a new :class:`File` object connected to the descriptor
        """
        fp = object.__new__(cls)  # bypass __init__
        fp._rbuf = StringIO()
        fp.encoding = None
        fp.mode = mode
        fp._fileno = fd
        fp._closed = False

        cls._add_flags(fd, cls._mode_to_flags(mode))
        fp._set_up_waiting()

        return fp

    def close(self):
        "close the file, and its underlying descriptor"
        self._closed = True
        _osclose(self._fileno)

    @property
    def closed(self):
        "return True if the file descriptor is closed"
        return self._closed

    def fileno(self):
        "get the file descriptor integer"
        return self._fileno

    def flush(self):
        """flush buffered writes to disk immediately

        this is provided for compatibility -- this class does no write
        buffering itself, so it is a no-op
        """
        return None

    def isatty(self):
        "return whether the file is connected to a tty or not"
        try:
            return os.isatty(self._fileno)
        except OSError, e:
            raise IOError(*e.args)

    def seek(self, position, modifier=0):
        """move the cursor on the file descriptor to a different location

        :param position:
            an integer offset from the location indicated by the modifier
        :type position: int
        :param modifier:
            an indicator of how to find the seek location.

            - ``os.SEEK_SET`` means start from the beginning of the file
            - ``os.SEEK_CUR`` means start wherever the cursor already is
            - ``os.SEEK_END`` means start from the end of the file

            the default is ``os.SEEK_SET``
        """
        os.lseek(self._fileno, position, modifier)

        # clear out the buffer
        buf = self._rbuf
        buf.seek(0)
        buf.truncate()

    def tell(self):
        "get the file descriptor's position relative to the file's beginning"
        with _fdopen(os.dup(self._fileno)) as fp:
            return fp.tell()


class _StdIOFile(FileBase):
    def __init__(self, fd):
        super(_StdIOFile, self).__init__()
        self._fileno = fd
        self._readable = util.Event()
        self._writable = util.Event()

    def _on_readable(self):
        self._readable.set()
        self._readable.clear()

    def _on_writable(self):
        self._writable.set()
        self._writable.clear()

    def _read_chunk(self, size):
        onr = self._on_readable
        reg = scheduler._register_fd(self._fileno, onr, None)
        try:
            self._readable.wait()
        finally:
            scheduler._unregister_fd(self._fileno, onr, None, reg)

        if scheduler.state.interrupted:
            raise IOError(errno.EINTR, "interrupted system call")

        return _read(self._fileno, size)

    def _write_chunk(self, data):
        onw = self._on_writable
        reg = scheduler._register_fd(self._fileno, None, onw)
        try:
            self._writable.wait()
        finally:
            scheduler._unregister_fd(self._fileno, None, onw, reg)

        if scheduler.state.interrupted:
            raise IOError(errno.EINTR, "interrupted system call")

        return _write(self._fileno, data)

    def close(self):
        _osclose(self._fileno)

    def fileno(self):
        return self._fileno

    def flush(self):
        return None

    def isatty(self):
        try:
            return os.isatty(self._fileno)
        except OSError, exc:
            raise IOError(*exc.args)

    def seek(self, position, modifier=0):
        raise IOError(errno.ESPIPE, "Illegal seek")

    def tell(self):
        raise IOError(errno.ESPIPE, "Illegal seek")


stdin = _StdIOFile(getattr(sys.stdin, "fileno", lambda: 0)())
"a non-blocking file object for cooperative stdin reading"

stdout = _StdIOFile(getattr(sys.stdout, "fileno", lambda: 1)())
"a non-blocking file object for cooperative stdout writing"

stderr = _StdIOFile(getattr(sys.stderr, "fileno", lambda: 2)())
"a non-blocking file object for cooperative stderr writing"



from __future__ import absolute_import

import errno
import functools
import time

from .. import compat, scheduler


__all__ = ["wait_fds"]


def wait_fds(fd_events, inmask=1, outmask=2, timeout=None):
    """wait for the first of a number of file descriptors to have activity

    .. note:: this method can block

        it will return once there is relevant activity on the file descriptors,
        or the timeout expires

    :param fd_events:
        two-tuples, each one a file descriptor and a mask made up of the inmask
        and/or the outmask bitwise-ORd together
    :type fd_events: list
    :param inmask: the mask to use for readable events (default 1)
    :type inmask: int
    :param outmask: the mask to use for writable events (default 2)
    :type outmask: int
    :param timeout:
        the maximum time to wait before raising an exception (default None)
    :type timeout: int, float or None

    :returns:
        a list of two-tuples, each is a file descriptor and an event mask (made
        up of inmask and/or outmask bitwise-ORd together) representing readable
        and writable events
    """
    current = compat.getcurrent()
    activated = {}
    poll_regs = {}
    callback_refs = {}

    def activate(fd, event):
        if not activated and timeout != 0:
            # this is the first invocation of `activated` for a blocking
            # `wait_fds` call, so re-schedule the blocked coroutine
            scheduler.schedule(current)

            # if there was a timeout then also have to pull
            # the coroutine from the timed_paused structure
            if timeout:
                scheduler._remove_timer(waketime, current)

        # in any case, set the event information
        activated.setdefault(fd, 0)
        activated[fd] |= event

    for fd, events in fd_events:
        readable = None
        writable = None
        if events & inmask:
            readable = functools.partial(activate, fd, inmask)
        if events & outmask:
            writable = functools.partial(activate, fd, outmask)

        callback_refs[fd] = (readable, writable)
        poll_regs[fd] = scheduler._register_fd(fd, readable, writable)

    if timeout:
        # real timeout value, schedule ourself `timeout` seconds in the future
        waketime = time.time() + timeout
        scheduler.pause_until(waketime)
    elif timeout == 0:
        # timeout == 0, only pause for 1 loop iteration
        scheduler.pause()
    else:
        # timeout is None, it's up to _hit_poller->activate to bring us back
        scheduler.state.mainloop.switch()

    for fd, reg in poll_regs.iteritems():
        readable, writable = callback_refs[fd]
        scheduler._unregister_fd(fd, readable, writable, reg)

    if scheduler.state.interrupted:
        raise IOError(errno.EINTR, "interrupted system call")

    return activated.items()



from __future__ import absolute_import, with_statement

import socket

from ..emulation import patched

exception = patched("dns.exception")
rdatatype = patched("dns.rdatatype")
resolver = patched("dns.resolver")
reversename = patched("dns.reversename")

DNS_CACHE_SIZE = 512


class LRU(object):
    def __init__(self, size):
        self._keyed = {}
        self._head = None
        self.size = size

    def _set_head(self, item):
        head = self._head
        self._head = item
        if head:
            item.prev = head.prev or head
            item.prev.next = item
            item.next = head
            head.prev = item
        else:
            item.prev = item.next = None

    def _remove_item(self, item, from_keyed=True):
        if item.prev:
            item.prev.next = item.next
        if item.next:
            item.next.prev = item.prev
        if from_keyed:
            self._keyed.pop(item.key)

    def _purge(self):
        while len(self._keyed) > self.size:
            item = self._head.prev
            self._remove_item(item)

    def __contains__(self, key):
        return key in self._keyed

    def __getitem__(self, key):
        item = self._keyed[key]
        self._remove_item(item, from_keyed=False)
        self._set_head(item)
        return item.value

    def __setitem__(self, key, value):
        old_item = self._keyed.get(key, None)
        if old_item:
            self._remove_item(old_item)

        item = LRUItem(key, value)
        self._set_head(item)
        self._keyed[key] = item
        self._purge()


class LRUItem(object):
    __slots__ = ["key", "value", "next", "prev"]

    def __init__(self, key, value):
        self.key = key
        self.value = value


etc_hosts_loaded = False
etc_hosts = {}
cache = LRU(DNS_CACHE_SIZE)
resolver_obj = None


def build_resolver(filename="/etc/resolv.conf"):
    if resolver_obj is None:
        globals()['resolver_obj'] = resolver.Resolver(filename)


def load_etc_hosts(filename="/etc/hosts"):
    with open(filename) as fp:
        for line in fp:
            try:
                if (c for c in line if not c.isspace()).next() == '#':
                    continue
            except StopIteration:
                continue

            parts = line.split()
            if '::' in parts[0]:
                # not handling ipv6 yet
                continue
            for part in parts[1:]:
                etc_hosts.setdefault(part, []).append(parts[0])
    globals()['etc_hosts_loaded'] = True


def is_ipv4(name):
    try:
        socket.inet_pton(socket.AF_INET, name)
    except socket.error:
        return False
    return True


def resolve(name):
    build_resolver()

    if name is None:
        return ["127.0.0.1"]

    if is_ipv4(name):
        return [name]

    if not etc_hosts_loaded:
        load_etc_hosts()

    # if it's in /etc/hosts then use that
    if name in etc_hosts:
        return etc_hosts[name][:]

    # return it out of the cache if it's there
    if name in cache:
        return cache[name][:]

    results = [x.to_text() for x in resolver_obj.query(name)]
    cache[name] = results[:]
    return results



from __future__ import absolute_import

from . import descriptor, files, ipc, sockets, ssl


__all__ = ["Socket", "File", "pipe", "stdin", "stdout", "stderr", "wait_fds",
        "SSLSocket", "wrap_socket"]


File = files.File
stdin = files.stdin
stdout = files.stdout
stderr = files.stderr

pipe = ipc.pipe

Socket = sockets.Socket

wait_fds = descriptor.wait_fds

wrap_socket = SSLSocket = ssl.SSLSocket



from __future__ import absolute_import

import socket
try:
    from . import dns
except ImportError:
    dns = None

from .. import io


def green_socketpair(*args, **kwargs):
    a, b = io.sockets._socketpair(*args, **kwargs)
    return io.Socket(fromsock=a), io.Socket(fromsock=b)


def create_connection(address, timeout=socket._GLOBAL_DEFAULT_TIMEOUT,
                      source_address=None):
    if dns:
        getaddrinfo = dns.getaddrinfo
    else:
        getaddrinfo = socket.getaddrinfo
    err = None
    for res in getaddrinfo(address[0], address[1], 0, socket.SOCK_STREAM):
        af, socktype, proto, canonname, sa = res
        sock = None
        try:
            sock = io.Socket(af, socktype, proto)
            if timeout is not socket._GLOBAL_DEFAULT_TIMEOUT:
                sock.settimeout(timeout)
            if source_address:
                sock.bind(source_address)
            sock.connect(sa)
            return sock
        except socket.error, exc:
            err = exc

    if err is not None:
        raise err
    else:
        raise socket.error("getaddrinfo returns an empty list")


patchers = {
    'socket': io.Socket,
    'SocketType': io.Socket,
    'socketpair': green_socketpair,
    'create_connection': create_connection,
    'fromfd': io.sockets.socket_fromfd,
}



try:
    from greenlet import greenlet
    try:
        from greenlet import GreenletExit
    except ImportError, error:
        try:
            GreenletExit = greenlet.GreenletExit
        except AttributeError:
            raise error
except ImportError, error:
    try:
        from py.magic import greenlet
        GreenletExit = greenlet.GreenletExit
    except ImportError:
        # suggest standalone greenlet, not the old py.magic.greenlet
        raise error


__all__ = ["main_greenlet", "getcurrent", "GreenletExit"]

getcurrent = greenlet.getcurrent

# it's conceivable that we might not be in the main greenlet at import time,
# so chase the parent tree until we get to it
main_greenlet = getcurrent()
while main_greenlet.parent:
    main_greenlet = main_greenlet.parent



from __future__ import absolute_import

import errno

try:
    import zmq
    from ..ext import zmq as gzmq
except ImportError:
    zmq = gzmq = None

try:
    import zmq.backend as zback
except ImportError:
    zback = None

try:
    import zmq.sugar as zs
except ImportError:
    zs = None


if zmq:
    original_backend_context = zmq.backend.Context
    original_backend_socket = zmq.backend.Socket

    class ZMQContextGreenifier(object):
        def socket(self, sock_type):
            return self._SOCK_CLASS(self, sock_type)

    class ZMQSocketGreenifier(object):
        def send(self, msg, flags=0):
            if flags & zmq.NOBLOCK:
                return super(ZMQSocketGreenifier, self).send(msg, flags)
            flags |= zmq.NOBLOCK

            while 1:
                try:
                    return super(ZMQSocketGreenifier, self).send(msg, flags)
                except zmq.ZMQError, exc:
                    if exc.errno != errno.EAGAIN:
                        raise
                    gzmq.wait_socks([(self, 2)])

        def recv(self, flags=0, copy=True, track=False):
            if flags & zmq.NOBLOCK:
                return super(ZMQSocketGreenifier, self).recv(
                    flags, copy, track)
            flags |= zmq.NOBLOCK

            while 1:
                try:
                    return super(ZMQSocketGreenifier, self).recv(
                        flags, copy, track)
                except zmq.ZMQError, exc:
                    if exc.errno != errno.EAGAIN:
                        raise
                    gzmq.wait_socks([(self, 1)])

    class ZMQBackendSocket(ZMQSocketGreenifier, original_backend_socket):
        pass

    class ZMQBackendContext(ZMQContextGreenifier, original_backend_context):
        _SOCK_CLASS = ZMQBackendSocket

    class ZMQPoller(object):
        def __init__(self):
            self._registry = {}

        def register(self, socket, flags=zmq.POLLIN | zmq.POLLOUT):
            self._registry[socket] = flags

        modify = register

        def unregister(self, socket):
            del self._registry[socket]

        def poll(self, timeout=None):
            return gzmq.wait_socks(self._registry.items(),
                                   inmask=zmq.POLLIN, outmask=zmq.POLLOUT,
                                   timeout=timeout)

    def zmq_poll(sockets, timeout=-1):
        if timeout < 0:
            timeout = None
        return gzmq.wait_socks(sockets,
                               inmask=zmq.POLLIN, outmask=zmq.POLLOUT,
                               timeout=timeout)

if zs:
    original_sugar_context = zs.Context
    original_sugar_socket = zs.Socket

    class ZMQSugarSocket(ZMQSocketGreenifier, original_sugar_socket):
        pass

    class ZMQSugarContext(ZMQContextGreenifier, original_sugar_context):
        _SOCK_CLASS = ZMQSugarSocket


patchers = {}
backend_patchers = {}
sugar_patchers = {}
sugar_context_patchers = {}
sugar_socket_patchers = {}
sugar_poll_patchers = {}

if zmq:
    patchers['Context'] = ZMQBackendContext
    patchers['Socket'] = ZMQBackendSocket
    patchers['zmq_poll'] = zmq_poll
    patchers['Poller'] = ZMQPoller

if zback:
    backend_patchers.update(patchers)

if zs:
    patchers['Context'] = ZMQSugarContext
    patchers['Socket'] = ZMQSugarSocket
    patchers['Poller'] = ZMQPoller
    sugar_patchers.update(patchers)
    #sugar_context_patchers['Context'] = ZMQSugarContext
    sugar_socket_patchers['Socket'] = ZMQSugarSocket
    sugar_poll_patchers['Poller'] = ZMQPoller
